import java.io.FileOutputStream;

import de.bioinf.appl.Appl;
import de.bioinf.appl.ApplParam;
import de.bioinf.utils.BioinfException;
import de.bioinf.utils.Logger;
import de.bioinf.utils.Matcher;
import de.bioinf.utils.Utils;

/**
 * Class for installing this software. A fundamental step is to create shell scripts in a specified location.
 * 
 * @author ng
 */
public class Install extends Appl {
	public static final String APPL_ID = "install";

	public Install(String[] args) throws BioinfException {
		super(APPL_ID, args, false);
	}

	public static void main(String[] args) throws BioinfException {
		new Install(args).run();
	}

	@Override
	public void run() throws BioinfException {
		try {
			String bindir = ApplParam.get("-d", Appl.cmdparams, (String) null);
			if (bindir == null) {
				System.out.println("Missing parameter -d!");
				Appl.stop();
			}
			createShellScripts(bindir);
		} catch (Exception e) {
			Logger.error(e);
		}
	}

	/**
	 * Create shell scripts for easy calls. First: -depending on the operating system- the templates bi_temples.bat or
	 * bi_template.sh are read from the jar archive. Second: for every class in the archive that is a subclass of Appl a
	 * shell script will be created from the template.
	 */
	public void createShellScripts(String bindir) throws BioinfException {
		// Assume the first classpath element is our jar archive!
		String jarname = Utils.getClassPathElements().get(0);
		if ("/".indexOf(jarname.charAt(0)) < 0) {
			jarname = Utils.getHomeDirectory() + "/" + jarname;
		}

		String osname = System.getProperty("os.name");
		String suffix = osname.startsWith("Windows") ? ".bat" : ".sh";
		String tmpl = new String(Utils.getClassPathContent("scripts/bi_template" + suffix));

		try {
			String[] names = Utils.getClassPathFiles(new Matcher("de/bioinf/appl/*.class"));
			for (String name : names) {
				try {
					// Convert filename to classname
					name = name.replace('/', '.');
					name = name.substring(0, name.length() - 6);
					Class c = Class.forName(name);
					if (c.getSuperclass() == Appl.class) {
						String applname = (String) c.getField("APPL_ID").get(c);

						// Replace placeholders
						String script = tmpl.replace("JARNAME", jarname);
						script = script.replace("APPLCLASS", c.getName());

						String filename = bindir + "/bi_" + applname + suffix;

						// Write to file in bindir
						FileOutputStream out = new FileOutputStream(filename);
						out.write(script.getBytes(), 0, script.length());
						out.close();
						out.flush();

						// Change execution mode on Unix
						if (!osname.equals("Windows")) {
							Process p = Runtime.getRuntime().exec(String.format("chmod a+rx %s", filename));
							if (p.waitFor() != 0) {
								Logger.error(String.format("Cannot change execution mode for %s!", filename));
							}
						}
					}
				} catch (Exception e) {
					Logger.error(BioinfException.convert(e).addMessage(
							String.format("Cannot create script for %s!", name)));
				}
			}
		} catch (Exception e) {
			Logger.error(e);
		}
	}
}
