package de.bioinf.appl;

import java.io.File;
import java.util.ArrayList;

import de.bioinf.utils.BioinfException;
import de.bioinf.utils.Logger;
import de.bioinf.utils.Utils;

/**
 * The base class for applications in Bioinformatics.
 * 
 * <p>
 * Remark: for proper installation it is neccessary to add a field APPL_ID to every application, like:
 * <code>public static final String APPL_ID = "crazy_tool";</code>
 */
public abstract class Appl {
	/**
	 * Parameters from the command line.
	 */
	public static CommandParams cmdparams = null;
	/**
	 * Parameters from the application specific configuration file.
	 */
	public static ConfigParams cnfparams = null;

	/**
	 * The components added to this application.
	 */
	private static ArrayList<ApplComponent> components = new ArrayList<ApplComponent>();

	/**
	 * Prepare the application and load the application parameters.
	 */
	protected Appl(String applid, String[] args, boolean log2file) throws BioinfException {
		super();

		/*
		 * if (System.getProperty("os.name").contains("Mac")) { System.setProperty("apple.laf.useScreenMenuBar",
		 * "true"); }
		 */

		Environment.init();

		cmdparams = new CommandParams(args);
		cnfparams = new ConfigParams(applid);

		// Write help message
		for (String key : cmdparams.getKeys()) {
			if ("|-help|-?|".indexOf("|" + key + "|") >= 0) {
				byte[] help = Utils.getClassPathContent("help/" + applid + ".help");
				System.out.println(help != null ? new String(help) : "No help available!");
				System.exit(0);
			}
		}

		Logger.setTrace(ApplParam.isKey("-trace", Appl.cmdparams));

		// Create the users configuration directory
		try {
			File f = new File(Utils.getConfDirectory());
			if (!f.exists()) {
				f.mkdir();
			}
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage(
					"Cannot create configuration directory '" + Utils.getConfDirectory() + "'!");
		}

		if (log2file) {
			Logger.redirect(ApplParam.get("-log", cmdparams, Utils.getConfDirectory() + "/" + applid + ".log"));
		}
	}

	/**
	 * Add special components to the application, so in the end (calling stop()), the components are automatically
	 * called to prepare their parameters for saving.
	 */
	public static void add(ApplComponent comp) {
		components.add(comp);
	}

	/**
	 * Stops the application (after saving the parameters of the registered components).
	 */
	public static void stop() {
		try {
			for (ApplComponent comp : components) {
				try {
					comp.prepareParams();
				} catch (Exception e) {
					Logger.error(e.getMessage());
				}
			}
			cnfparams.save();
			System.exit(0);
		} catch (Exception e) {
			System.exit(1);
		}
	}

	/**
	 * The "main" method to start the application (e.g. after processing command parameters).
	 */
	public abstract void run() throws BioinfException;
}
