package de.bioinf.appl;

import java.awt.Color;
import java.awt.Dimension;
import java.util.StringTokenizer;

import de.bioinf.utils.Logger;
import de.bioinf.utils.Params;

/**
 * This is a helper class for the application parameters. Mostly the parameters
 * are converted from the storage representation (as String) to the useful
 * internal value types (using get()) and back again (using set()).
 * 
 * <p>
 * To use this ApplParam mechanisms it is neccessary to create a class derived
 * from Appl and then just type:
 * <code>double d = ApplParam.get("-d", Appl.cmdparams, 0.01);</code> or
 * <code>Dimension d = ApplParam.get("size", Appl.cnfparams, new Dimension(100,100));</code>
 * 
 * @see CommandParams
 * @see ConfigParams
 * 
 * @author ng
 */
public class ApplParam {
	private ApplParam() {
	}

	// int

	public static int get(String key, Params<String, String> params, int defval) {
		String sval = params.getValue(key);
		if (sval != null) {
			try {
				return Integer.parseInt(sval);
			} catch (Exception e) {
				error(key, sval);
			}
		}
		return defval;
	}

	public static void set(String key, Params<String, String> params, int value) {
		params.put(key, Integer.toString(value));
	}

	// double

	public static double get(String key, Params<String, String> params, double defval) {
		String sval = params.getValue(key);
		if (sval != null) {
			try {
				return Double.parseDouble(sval);
			} catch (Exception e) {
				error(key, sval);
			}
		}
		return defval;
	}

	public static void set(String key, Params<String, String> params, double value) {
		params.put(key, Double.toString(value));
	}

	// String

	public static String get(String key, Params<String, String> params, String defval) {
		String sval = params.getValue(key);
		return sval != null ? sval : defval;
	}

	public static void set(String key, Params<String, String> params, String value) {
		params.put(key, value);
	}

	// boolean

	public static boolean get(String key, Params<String, String> params, boolean defval) {
		String sval = params.getValue(key);
		if (sval != null) {
			try {
				return Boolean.parseBoolean(sval);
			} catch (Exception e) {
				error(key, sval);
			}
		}
		return defval;
	}

	public static void set(String key, Params<String, String> params, boolean value) {
		params.put(key, Boolean.toString(value));
	}

	// Dimension

	public static Dimension get(String key, Params<String, String> params, Dimension defval) {
		String sval = params.getValue(key);
		if (sval != null) {
			try {
				int index = sval.indexOf("x");
				return new Dimension(Integer.parseInt(sval.substring(0, index)), Integer.parseInt(sval
						.substring(index + 1)));
			} catch (Exception e) {
				error(key, sval);
			}
		}
		return defval;
	}

	public static void set(String key, Params<String, String> params, Dimension value) {
		params.put(key, value.width + "x" + value.height);
	}

	// Color

	public static Color get(String key, Params<String, String> params, Color defval) {
		String sval = params.getValue(key);
		if (sval != null) {
			try {
				StringTokenizer tok = new StringTokenizer(sval, ",");
				int r = Integer.parseInt(tok.nextToken().trim());
				int g = Integer.parseInt(tok.nextToken().trim());
				int b = Integer.parseInt(tok.nextToken().trim());
				return new Color(r, g, b);
			} catch (Exception e) {
				error(key, sval);
			}
		}
		return defval;
	}

	public static void set(String key, Params<String, String> params, Color value) {
		params.put(key, String.format("%d,%d,%d", value.getRed(), value.getGreen(), value.getBlue()));
	}

	// Misc

	public static boolean isKey(String key, Params<String, String> params) {
		return params.isKey(key);
	}

	private static void error(String key, String sval) {
		Logger.error(String.format("Parameter %s has wrong value %s!", key, sval));
	}
}
