package de.bioinf.appl;

import de.bioinf.base.Codon;
import de.bioinf.base.CodonSequence;
import de.bioinf.base.Nucleotid;
import de.bioinf.base.NucleotidSequence;
import de.bioinf.base.NucleotidSequences;
import de.bioinf.utils.BioinfException;
import de.bioinf.utils.Logger;
import de.bioinf.utils.Source;

/**
 * Analyzer for 1 and 3 nucleotid sequences.
 * 
 * @author ng
 */
public class Sequence3Analyzer extends Appl {
	private String p_filename = null;

	public static void main(String[] args) {
		try {
			new Sequence3Analyzer(args).run();
		} catch (Exception e) {
			Logger.error(BioinfException.convert(e).toString());
		}
	}

	public Sequence3Analyzer(String[] args) throws BioinfException {
		super("sequence", args, false);
		evalParams();
	}

	public void evalParams() throws BioinfException {
		p_filename = ApplParam.get("-f", Appl.cmdparams, (String) null);
		BioinfException.fire(p_filename != null, "Missing parameter -f!");
	}

	@Override
	public void run() throws BioinfException {
		NucleotidSequences nseqs = NucleotidSequences.loadFasta(new Source(p_filename));
		for (int i = 0; i < nseqs.size(); i++) {
			NucleotidSequence nseq = nseqs.get(i);
			System.out.println(nseq.getHeader().toString());

			int[] counters = countNucleotids(nseq);
			for (int k = 0; k < counters.length; k++) {
				System.out.println(Nucleotid.getDnaNucleotid(k) + " = " + counters[k]);
			}

			counters = countCodons(new CodonSequence(new CodonSequence(nseq)));
			for (int k = 0; k < counters.length; k++) {
				System.out.println(Codon.getCodon(k) + " = " + counters[k]);
			}
		}
	}

	// **
	// Private
	// **

	/**
	 * Count the nucleotid types in the sequence
	 */
	private int[] countNucleotids(NucleotidSequence seq) {
		int[] counters = new int[Nucleotid.COUNT];
		for (int i = 0; i < counters.length; i++) {
			counters[i] = 0;
		}
		for (int i = 0; i < seq.length(); i++) {
			char nuc = seq.getNucleotid(i);
			if (Nucleotid.isValid(nuc)) {
				counters[Nucleotid.getIndex(seq.getNucleotid(i))]++;
			}
		}
		return counters;
	}

	/**
	 * Count the codon types in the sequences
	 */
	private int[] countCodons(CodonSequence seq) {
		int[] counters = new int[Codon.COUNT];
		for (int i = 0; i < counters.length; i++) {
			counters[i] = 0;
		}
		for (int i = 0; i < seq.length(); i++) {
			Codon codon = seq.getCodon(i);
			if (codon.isValid()) {
				counters[seq.getCodon(i).getIndex()]++;
			}
		}
		return counters;
	}
}
