package de.bioinf.appl.distg;

import java.util.ArrayList;
import java.util.HashMap;

import de.bioinf.base.FastaSeq;
import de.bioinf.utils.BioinfException;
import de.bioinf.utils.Counter;
import de.bioinf.utils.IntValueRange;
import de.bioinf.utils.LineReader;
import de.bioinf.utils.LineWriter;
import de.bioinf.utils.SequenceSelector;
import de.bioinf.utils.Source;

/**
 * A class to summarize and abstract the counts of the sequence length of a fasta file.
 * 
 * @author ng
 */
public class DistGValues {
	protected Source source;
	protected int[] counts = null;

	private DistGValues(Source source, int[] counts) {
		super();
		this.source = source;
		this.counts = counts;
	}

	public boolean equals(Object o) {
		if (o == null || !(o instanceof DistGValues)) {
			return false;
		}
		return source.equals(((DistGValues) o).source);
	}

	/**
	 * Saves this counts into a file.
	 */
	public void save(String filename, DistGParams params) throws BioinfException {
		LineWriter writer = null;
		try {
			writer = new LineWriter(filename);
			for (int i = params.countRange.getMinVal(); i + params.bucketSize < params.countRange.getMaxVal(); i += params.bucketSize) {
				int sum = 0;
				for (int j = 0; j < params.bucketSize; j++) {
					sum += counts.length > i + j ? counts[i + j] : 0;
				}
				writer.writeLine(i + " " + sum);
			}
		} catch (Exception e) {
			BioinfException.convert(e);
		} finally {
			if (writer != null) {
				try {
					writer.close();
				} catch (Exception e2) {
					throw BioinfException.convert(e2);
				}
			}
		}
	}

	/**
	 * Does any counter exist inside the range with a count > 0?
	 * 
	 * Note: If not the display may make problems, because the calculated y range for the counter representation may
	 * only be in 0..0!
	 */
	public boolean isValid(IntValueRange range) {
		for (int i = range.getMinVal(); i < range.getMaxVal() && i < counts.length; i++) {
			if (counts[i] > 0) {
				return true;
			}
		}
		return false;
	}

	/**
	 * The same as isValid(range), but for a list of values
	 */
	public static boolean isValid(ArrayList<DistGValues> list, IntValueRange range) {
		for (DistGValues values : list) {
			if (values.isValid(range)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns the range of all values (starting with 0!)
	 */
	public static IntValueRange getTotalRange(ArrayList<DistGValues> list) {
		IntValueRange range = new IntValueRange(0, 0);
		for (DistGValues values : list) {
			range.add(values.counts.length);
		}
		return range;
	}

	/**
	 * Looks if the source is already in the list.
	 */
	public static boolean isLoaded(Source source, ArrayList<DistGValues> list) {
		for (DistGValues values : list) {
			if (values.source.equals(source)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Loads and creates an object - and recognize only sequences of the desired length range
	 */
	public static DistGValues load(Source source, DistGParams params) throws BioinfException {
		LineReader reader = null;
		try {
			SequenceSelector selector = new SequenceSelector() {
				@Override
				public boolean isOk(String header) throws BioinfException {
					return true;
				}
			};

			reader = source.getReader(true);
			FastaSeq seq = null;
			int max = 0;
			HashMap<Integer, Counter> map = new HashMap<Integer, Counter>();
			while ((seq = FastaSeq.nextSeq(reader, selector)) != null) {
				int len = seq.seq.length();
				if (len >= params.bucketCell) {
					len = params.bucketCell;
				}
				if (params.lenRange.isInside(len)) {
					Counter counter = map.get(len);
					if (counter != null) {
						counter.inc();
					} else {
						map.put(len, new Counter(1));
					}
					max = Math.max(len + 1, max);
				}
			}
			BioinfException.fire(map.size() > 0, "No valid fasta data!");

			int[] counts = new int[max];
			for (int i = 0; i < max; i++) {
				Counter counter = map.get(i);
				counts[i] = counter != null ? counter.getValue() : 0;
			}

			return new DistGValues(source, counts);
		} catch (Exception e) {
			throw BioinfException.convert(e);
		} finally {
			reader.close();
		}
	}

	/**
	 * Saves by summarizing the counts of a set of values into one file
	 */
	public static void save(ArrayList<DistGValues> list, String filename, DistGParams params) throws BioinfException {
		LineWriter writer = null;
		try {
			writer = new LineWriter(filename);
			for (int i = params.countRange.getMinVal(); i + params.bucketSize < params.countRange.getMaxVal(); i += params.bucketSize) {
				int sum = 0;
				for (int j = 0; j < params.bucketSize; j++) {
					for (DistGValues values : list) {
						sum += values.counts.length > i + j ? values.counts[i + j] : 0;
					}
				}
				writer.writeLine(i + " " + sum);
			}
		} catch (Exception e) {
			BioinfException.convert(e);
		} finally {
			if (writer != null) {
				try {
					writer.close();
				} catch (Exception e2) {
					throw BioinfException.convert(e2);
				}
			}
		}
	}
}
