package de.bioinf.base;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.Hashtable;

import de.bioinf.utils.BioinfException;
import de.bioinf.utils.LineReader;
import de.bioinf.utils.Logger;

/**
 * This class provides information about the mapping from codons to amino acids
 * and vice versa.
 * 
 * @author ng
 */
public class CodonMap {
	private Hashtable<Codon, AminoAcid> codon2acid = new Hashtable<Codon, AminoAcid>();
	private Hashtable<AminoAcid, ArrayList<Codon>> acid2codons = new Hashtable<AminoAcid, ArrayList<Codon>>();

	/**
	 * The default map read from conf/codon.map
	 */
	private static CodonMap DEFAULT = null;

	/**
	 * Creates the default mapping from the Bioinf.jar archive.
	 */
	private CodonMap() throws BioinfException {
		super();
		init();
	}

	public CodonMap(String filename) throws BioinfException {
		super();
		init(filename);
	}

	public void add(AminoAcid acid, Codon codon) {
		codon2acid.put(codon, acid);
		ArrayList<Codon> codons = acid2codons.get(acid);
		if (codons == null) {
			codons = new ArrayList<Codon>();
			acid2codons.put(acid, codons);
		}
		codons.add(codon);
	}

	/**
	 * Returns the amino acid for the codon.
	 */
	public AminoAcid getAcid(Codon c) {
		return codon2acid.get(c);
	}

	/**
	 * Returns all codons that are "connected" to an amino acid (in that order
	 * read from the file).
	 */
	public ArrayList<Codon> getCodons(AminoAcid acid) {
		return acid2codons.get(acid);
	}

	public static CodonMap getDefaultMap() {
		return DEFAULT;
	}

	public String toString() {
		return codon2acid.toString();
	}

	// **
	// Private
	// **

	/**
	 * Initializes this mapping using the file.
	 */
	private void init(String filename) throws BioinfException {
		fill(new LineReader(filename));
	}

	private void init() throws BioinfException {
		fill(new LineReader(new StringReader(getDefaultData())));
	}

	private void fill(LineReader reader) throws BioinfException {
		String line = null;
		try {
			while ((line = reader.readLine()) != null) {
				int index = line.indexOf('=');
				if (index != -1) {
					Codon codon = Codon.create(line.substring(0, index).trim());
					AminoAcid acid = AminoAcid.get(line.substring(index + 1).trim());
					add(acid, codon);
				}
			}
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage("Line '" + line + "' has error!");
		} finally {
			reader.close();
		}
	}

	private String getDefaultData() {
		StringBuffer buf = new StringBuffer();
		buf.append("TTT = Phe\n");
		buf.append("TTC = Phe\n");
		buf.append("TTA = Leu\n");
		buf.append("TTG = Leu\n");
		buf.append("TCT = Ser\n");
		buf.append("TCC = Ser\n");
		buf.append("TCA = Ser\n");
		buf.append("TCG = Ser\n");
		buf.append("TAT = Tyr\n");
		buf.append("TAC = Tyr\n");
		buf.append("TAA = Ter\n");
		buf.append("TAG = Ter\n");
		buf.append("TGT = Cys\n");
		buf.append("TGC = Cys\n");
		buf.append("TGA = Ter\n");
		buf.append("TGG = Trp\n");
		buf.append("CTT = Leu\n");
		buf.append("CTC = Leu\n");
		buf.append("CTA = Leu\n");
		buf.append("CTG = Leu\n");
		buf.append("CCT = Pro\n");
		buf.append("CCC = Pro\n");
		buf.append("CCA = Pro\n");
		buf.append("CCG = Pro\n");
		buf.append("CAT = His\n");
		buf.append("CAC = His\n");
		buf.append("CAA = Gln\n");
		buf.append("CAG = Gln\n");
		buf.append("CGT = Arg\n");
		buf.append("CGC = Arg\n");
		buf.append("CGA = Arg\n");
		buf.append("CGG = Arg\n");
		buf.append("ATT = Ile\n");
		buf.append("ATC = Ile\n");
		buf.append("ATA = Ile\n");
		buf.append("ATG = Met\n");
		buf.append("ACT = Thr\n");
		buf.append("ACC = Thr\n");
		buf.append("ACA = Thr\n");
		buf.append("ACG = Thr\n");
		buf.append("AAT = Asn\n");
		buf.append("AAC = Asn\n");
		buf.append("AAA = Lys\n");
		buf.append("AAG = Lys\n");
		buf.append("AGT = Ser\n");
		buf.append("AGC = Ser\n");
		buf.append("AGA = Arg\n");
		buf.append("AGG = Arg\n");
		buf.append("GTT = Val\n");
		buf.append("GTC = Val\n");
		buf.append("GTA = Val\n");
		buf.append("GTG = Val\n");
		buf.append("GCT = Ala\n");
		buf.append("GCC = Ala\n");
		buf.append("GCA = Ala\n");
		buf.append("GCG = Ala\n");
		buf.append("GAT = Asp\n");
		buf.append("GAC = Asp\n");
		buf.append("GAA = Glu\n");
		buf.append("GAG = Glu\n");
		buf.append("GGT = Gly\n");
		buf.append("GGC = Gly\n");
		buf.append("GGA = Gly\n");
		buf.append("GGG = Gly\n");
		return buf.toString();
	}

	static {
		try {
			DEFAULT = new CodonMap();
		} catch (Exception e) {
			Logger.error("Cannot create default map", e);

		}
	}
}
