package de.bioinf.base;

import java.util.StringTokenizer;

public class NucleotidSequence {
	protected SequenceHeader header = null;
	protected String sequence = null;

	public NucleotidSequence() {
		this(null, null);
	}

	public NucleotidSequence(String sequence) {
		this(null, sequence);
	}

	public NucleotidSequence(SequenceHeader header, String sequence) {
		super();
		this.header = header;
		this.sequence = sequence != null ? sequence : new String();
	}

	/**
	 * Returns the number of occurances of nucleotids in the sequence
	 */
	public int count(char nucleotid) {
		int count = 0;
		for (int i = 0; i < sequence.length(); i++) {
			char c = sequence.charAt(i);
			if (c == nucleotid) {
				count++;
			}
		}
		return count;
	}

	public int count(char[] nucs) {
		int count = 0;
		String s = new String(nucs);
		for (int i = 0; i < sequence.length(); i++) {
			if (s.indexOf(sequence.charAt(i)) >= 0) {
				count++;
			}
		}
		return count;
	}

	/**
	 * Count the pattern occurences in the sequence. The pattern may have the
	 * wildcard characters "." and "_" for a single match and for a undefined
	 * length match (even length 0!)
	 */
	public int count(String pattern) {
		StringTokenizer tok = new StringTokenizer(pattern, "_.", true);
		// No wildcard used: avoid complex count!
		if (tok.countTokens() == 1) {
			return count(new NucleotidSequence(tok.nextToken()));
		}
		String[] tokens = new String[tok.countTokens()];
		for (int i = 0; i < tokens.length; i++) {
			tokens[i] = tok.nextToken();
		}

		int count = 0;
		boolean multi = false;
		for (int pos = 0; pos < sequence.length();) {
			int newpos = 0;
			for (String token : tokens) {
				if ("_".equals(token)) {
					multi = true;
					continue;
				} else if (".".equals(token)) {
					pos++;
					continue;
				}
				if (multi) {
					newpos = sequence.indexOf(token, pos);
					multi = false;
				} else {
					newpos = sequence.indexOf(token, pos);
					if (newpos != pos) {
						newpos = -1;
					}
				}
				if (newpos == -1) {
					break;
				} else {
					pos = newpos + token.length();
				}
			}
			if (newpos != -1) {
				count++;
			} else {
				pos++;
			}
		}

		return count;
	}

	/**
	 * Returns the number of occurances of the sequence.
	 */
	public int count(NucleotidSequence seq) {
		int count = 0;
		int index = 0;
		while (index >= 0 && index < length()) {
			if ((index = sequence.indexOf(seq.sequence, index)) >= 0) {
				count++;
				index += seq.length();
			}
		}
		return count;
	}

	public boolean equals(Object o) {
		if (o == null || !(o instanceof NucleotidSequence)) {
			return false;
		}
		return sequence.equals(((NucleotidSequence) o).sequence);
	}

	/**
	 * Return the (first) position of nucleotid, or -1 if not found.
	 */
	public int find(char nuc) {
		return find(nuc, 0);
	}

	/**
	 * Returns the position (starting with position start), or -1 if not found.
	 * 
	 * @param nuc - The nucleotid searched for.
	 */
	public int find(char nuc, int start) {
		return sequence.indexOf(nuc, start);
	}

	/**
	 * Return the (first) position of sequence, or -1 if not found.
	 */
	public int find(NucleotidSequence seq) {
		return find(seq, 0);
	}

	/**
	 * Returns the position (starting with position start), or -1 if not found.
	 */
	public int find(NucleotidSequence seq, int start) {
		return sequence.indexOf(seq.sequence, start);
	}

	/**
	 * Returns the header (information line).
	 */
	public SequenceHeader getHeader() {
		return header;
	}

	/**
	 * Returns the nucleotid on position index
	 */
	public char getNucleotid(int index) {
		return sequence.charAt(index);
	}

	public int hashCode() {
		return sequence.hashCode();
	}

	/**
	 * Returns the length (number of nucleotids).
	 */
	public int length() {
		return sequence.length();
	}

	/**
	 * Returns a subsequence of nucleotids.
	 * 
	 * @param start - the starting index (starting with 0!)
	 * @param end - the ending index
	 */
	public NucleotidSequence subsequence(int start, int end) {
		return new NucleotidSequence(header, sequence.substring(start, end));
	}

	/**
	 * Returns the nucleotids in this sequence.
	 */
	public String toString() {
		return sequence;
	}
}
