package de.bioinf.base;

import java.util.ArrayList;

import de.bioinf.utils.BioinfException;
import de.bioinf.utils.LineReader;
import de.bioinf.utils.Logger;
import de.bioinf.utils.SequenceSelector;
import de.bioinf.utils.Source;

public class NucleotidSequences extends ArrayList<NucleotidSequence> {
	/**
	 * Loads all nucleotid sequences of a fasta file.
	 */
	public static NucleotidSequences loadFasta(Source source) throws BioinfException {
		return loadFasta(source, new SequenceSelector() {
			public boolean isOk(String line) throws BioinfException {
				return true;
			}
		});
	}

	/**
	 * Loads specific nucleotid sequences of a fasta file. A nucleotid sequence file must look like {">" info newline
	 * {data newline}}. That means an infoline is the first (starting with ">") followed by many lines of nucleotids.
	 * 
	 * @param selector - providing the possibility to select only specific sequences.
	 */
	public static NucleotidSequences loadFasta(Source source, SequenceSelector selector) throws BioinfException {
		Logger.info(String.format("Loading file %s", source.getLocation()));
		NucleotidSequences seqs = new NucleotidSequences();

		LineReader in = source.getReader(false);
		String line = null;
		int lineno = 0;

		try {
			SequenceHeader info = null;
			StringBuffer nucs = new StringBuffer(512);
			while ((line = in.readLine()) != null) {
				line = line.trim();
				lineno++;
				if (line.startsWith(">")) {
					if (info != null) {
						seqs.add(new NucleotidSequence(info, nucs.toString()));
						nucs = new StringBuffer(512);
					}
					info = selector.isOk(line) ? new SequenceHeader(line) : null;
				} else {
					if (info != null) {
						nucs.append(line.toUpperCase());
					}
				}
			}
			seqs.add(new NucleotidSequence(info, nucs.toString()));
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage(
					String.format("Error in line %d of file '%s'!", lineno, source.getLocation()));
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (Exception e) {
					Logger.error("Error closing file '" + source.getLocation() + "'!");
				}
			}
		}

		Logger.info("");
		return seqs;
	}
}
