package de.bioinf.ui;

import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.util.ArrayList;

import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.bioinf.utils.SelectionEvent;
import de.bioinf.utils.SelectionListener;

/**
 * This is a panel with 3 sliders representing the amount of red, green, blue parts in a color. On the right hand side
 * there is a panel that shows the actual color. It is possible to add listeners and act on selection immediately.
 * 
 * @author ng
 */
public class ColorSelector extends GridBagPanel implements InputComponent {
	// The internal graphical elements
	private JSlider g_rslider = null;
	private JSlider g_gslider = null;
	private JSlider g_bslider = null;
	private JPanel g_cpanel = null;

	// The values of the color parts
	private IntegerTextField g_rvalue = null;
	private IntegerTextField g_gvalue = null;
	private IntegerTextField g_bvalue = null;

	private Color color = new Color(0, 0, 0);
	private Color savedColor = null;

	private ArrayList<SelectionListener<ColorSelector>> listeners = new ArrayList<SelectionListener<ColorSelector>>();
	private boolean listenersActive = true;

	/**
	 * The color (slider) listener.
	 */
	abstract class CListener implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			g_cpanel.setBackground(color);
			if (listenersActive) {
				for (SelectionListener<ColorSelector> l : listeners) {
					l.selected(new SelectionEvent<ColorSelector>(ColorSelector.this));
				}
			}
		}
	}

	class RListener extends CListener {
		public void stateChanged(ChangeEvent e) {
			color = new Color(((JSlider) e.getSource()).getValue(), color.getGreen(), color.getBlue());
			g_rvalue.setText(Integer.toString(color.getRed()));
			super.stateChanged(e);
		}
	}

	class GListener extends CListener {
		public void stateChanged(ChangeEvent e) {
			color = new Color(color.getRed(), ((JSlider) e.getSource()).getValue(), color.getBlue());
			g_gvalue.setText(Integer.toString(color.getGreen()));
			super.stateChanged(e);
		}
	}

	class BListener extends CListener {
		public void stateChanged(ChangeEvent e) {
			color = new Color(color.getRed(), color.getGreen(), ((JSlider) e.getSource()).getValue());
			g_bvalue.setText(Integer.toString(color.getBlue()));
			super.stateChanged(e);
		}
	}

	abstract class ValueListener implements FocusListener {
		public void focusGained(FocusEvent e) {
		}
	}

	class RValueListener extends ValueListener {
		public void focusLost(FocusEvent e) {
			g_rslider.setValue(g_rvalue.getValue());
		}
	}

	class GValueListener extends ValueListener {
		public void focusLost(FocusEvent e) {
			g_gslider.setValue(g_gvalue.getValue());
		}
	}

	class BValueListener extends ValueListener {
		public void focusLost(FocusEvent e) {
			g_bslider.setValue(g_bvalue.getValue());
		}
	}

	/**
	 * Creates a selector with black as the default color.
	 */
	public ColorSelector() {
		this(Color.BLACK);
	}

	/**
	 * Creates a selector with the given color.
	 */
	public ColorSelector(Color c) {
		super();

		add(g_rslider = new JSlider(JSlider.VERTICAL, 0, 255, 0), GridBagConstraints.BOTH, 0, 0, 1, 1);
		g_rslider.setBackground(Color.RED);
		add(g_gslider = new JSlider(JSlider.VERTICAL, 0, 255, 0), GridBagConstraints.BOTH, 1, 0, 1, 1);
		g_gslider.setBackground(Color.GREEN);
		add(g_bslider = new JSlider(JSlider.VERTICAL, 0, 255, 0), GridBagConstraints.BOTH, 2, 0, 1, 1);
		g_bslider.setBackground(Color.BLUE);
		add(g_cpanel = new JPanel(), GridBagConstraints.BOTH, 3, 0, 1, 1);

		add(g_rvalue = new IntegerTextField(0, 255), GridBagConstraints.BOTH, 0, 1, 1, 1);
		add(g_gvalue = new IntegerTextField(0, 255), GridBagConstraints.BOTH, 1, 1, 1, 1);
		add(g_bvalue = new IntegerTextField(0, 255), GridBagConstraints.BOTH, 2, 1, 1, 1);

		setSelectedColor(c);

		// Add the listeners here, so NO actions are made so far!
		g_rslider.addChangeListener(new RListener());
		g_gslider.addChangeListener(new GListener());
		g_bslider.addChangeListener(new BListener());

		g_rvalue.addFocusListener(new RValueListener());
		g_gvalue.addFocusListener(new GValueListener());
		g_bvalue.addFocusListener(new BValueListener());
	}

	/**
	 * Add a listener to act immediately on selection.
	 */
	public void addSelectionListener(SelectionListener<ColorSelector> listener) {
		listeners.add(listener);
	}

	/**
	 * Returns the selected color.
	 */
	public Color getSelectedColor() {
		return color;
	}

	/**
	 * Sets the selected color, but prohibits the activation of all listeners
	 */
	public synchronized void setSelectedColor(Color c) {
		saveState();

		listenersActive = false;

		g_rslider.setValue(c.getRed());
		g_gslider.setValue(c.getGreen());
		g_bslider.setValue(c.getBlue());

		g_rvalue.setText(Integer.toString(c.getRed()));
		g_gvalue.setText(Integer.toString(c.getGreen()));
		g_bvalue.setText(Integer.toString(c.getBlue()));

		listenersActive = true;
	}

	public boolean isOk() {
		return true;
	}

	public void reset() {
		setSelectedColor(color = savedColor);
	}

	public void saveState() {
		savedColor = new Color(color.getRed(), color.getGreen(), color.getBlue());
	}
}
