package de.bioinf.ui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.LayoutManager;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.io.File;
import java.io.FileFilter;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashSet;

import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JViewport;

import de.bioinf.appl.Colors;
import de.bioinf.utils.Utils;

/**
 * Class to simplify file selections. It is more simple than dialogs like JFileChooser and FileDialog, but has the
 * options to select parent directories (and on Windows the drives) and enter a filename on bottom.
 * 
 * @author ng
 */
public class FileSelector extends JPanel implements InputComponent {
	private File selectedPath = null;
	private File savedPath = null;

	// Display hidden files
	private boolean hidden = false;

	private Header g_header = null;
	private ScrollList g_filelist = null;
	private Footer g_footer = null;

	private static final Color BACKGROUND = new Color(0xf8, 0xf8, 0xf8);

	/**
	 * The header line with the roots (if more than one) and the directory and parent directories.
	 */
	private class Header extends JPanel implements ActionListener {
		JComboBox g_dirs = null;

		public Header() {
			super(new BorderLayout());
			setBackground(BACKGROUND);
			JLabel label = new JLabel("Goto");
			label.setPreferredSize(new Dimension(60, 20));
			add(label, BorderLayout.WEST);
			add(g_dirs = new JComboBox(), BorderLayout.CENTER);
			g_dirs.setPreferredSize(new Dimension(100, 20));
		}

		public void init(File path) {
			if (path != null) {
				g_dirs.removeActionListener(this);
				g_dirs.removeAllItems();
				HashSet<File> h = new HashSet<File>();
				for (File file = path; file != null; file = file.getParentFile()) {
					h.add(file);
				}
				for (File file : File.listRoots()) {
					h.add(file);
				}
				File[] files = h.toArray(new File[h.size()]);
				Arrays.sort(files);
				for (File file : files) {
					g_dirs.addItem(file);
				}
				g_dirs.setSelectedItem(path);
				g_dirs.addActionListener(this);
			}
		}

		public void actionPerformed(ActionEvent e) {
			FileSelector.this.init((File) g_dirs.getSelectedItem());
		}
	}

	/**
	 * The list of files
	 */
	class FileList extends JPanel {
		File[] files = null;
		int lineheight = 0;

		class MListener implements MouseListener, MouseMotionListener {
			int index = -1;
			int selectedIndex = -1;

			public void mouseClicked(MouseEvent e) {
				int i = getIndex(e.getY());
				if (i < getComponentCount() / 3) {
					setForeground(selectedIndex, Color.BLACK);
					File file = files[i];
					if (file.isDirectory()) {
						FileSelector.this.init(file);
					} else if (file.isFile()) {
						FileSelector.this.setSelectedFile(file);
					}
					setForeground(selectedIndex = i, Colors.SELECTED_FG);
				}
			}

			public void mouseEntered(MouseEvent e) {
			}

			public void mouseExited(MouseEvent e) {
				setBackground(index, BACKGROUND);
			}

			public void mousePressed(MouseEvent e) {
			}

			public void mouseReleased(MouseEvent e) {
			}

			public void mouseMoved(MouseEvent e) {
				setBackground(index, BACKGROUND);
				setBackground(index = getIndex(e.getY()), Colors.SELECTED_BG);
			}

			public void mouseDragged(MouseEvent e) {
			}

			private int getIndex(int ypos) {
				return ypos / lineheight;
			}

			private void setBackground(int index, Color color) {
				if (index >= 0 && index < getComponentCount() / 3) {
					((JLabel) getComponent(index * 3)).setBackground(color);
					((JLabel) getComponent(index * 3 + 1)).setBackground(color);
					((JLabel) getComponent(index * 3 + 2)).setBackground(color);
				}
			}

			private void setForeground(int index, Color color) {
				if (index >= 0 && index < getComponentCount() / 3) {
					((JLabel) getComponent(index * 3)).setForeground(color);
					((JLabel) getComponent(index * 3 + 1)).setForeground(color);
					((JLabel) getComponent(index * 3 + 2)).setForeground(color);
				}
			}
		}

		LayoutManager layout = new LayoutManager() {
			int wdate = 0;
			int wsize = 0;
			int hgap = 1;

			public void addLayoutComponent(String name, Component comp) {
			}

			public void layoutContainer(Container parent) {
				if (lineheight == 0) {
					FontMetrics m = getGraphics().getFontMetrics();
					lineheight = m.getHeight() + hgap;
					wdate = m.stringWidth("9999-99-99 99:99") + 10;
					wsize = m.stringWidth("9999999999") + 10;
				}
				int width = parent.getSize().width;
				for (int i = 0; i < parent.getComponentCount() / 3; i++) {
					JLabel label = (JLabel) getComponent(i * 3);
					label.setLocation(5, i * lineheight);
					label.setSize(width - (wdate + wsize) - 10, lineheight);
					label = (JLabel) getComponent(i * 3 + 1);
					label.setLocation(width - (wdate + wsize) - 5, i * lineheight);
					label.setSize(wdate, lineheight);
					label = (JLabel) getComponent(i * 3 + 2);
					label.setLocation(width - wsize - 5, i * lineheight);
					label.setSize(wsize, lineheight);
				}
			}

			public Dimension minimumLayoutSize(Container parent) {
				return new Dimension(100, 100);
			}

			public Dimension preferredLayoutSize(Container parent) {
				return new Dimension(100, files.length * lineheight + hgap);
			}

			public void removeLayoutComponent(Component comp) {
			}
		};

		public FileList(File[] files) {
			super(null);
			setBackground(BACKGROUND);
			setLayout(layout);
			for (File file : this.files = files) {
				add(getLabel(file.getName()));
				add(getLabel(Utils.DATE_FORMAT.format(file.lastModified())));
				add(getLabel(file.isDirectory() ? "[DIR]" : Long.toString(file.length())));
			}
			MListener mlistener = new MListener();
			addMouseListener(mlistener);
			addMouseMotionListener(mlistener);
		}

		private JLabel getLabel(String label) {
			JLabel l = new JLabel(label);
			l.setBackground(BACKGROUND);
			l.setOpaque(true);
			return l;
		}
	}

	/**
	 * A view on the files in a directory. The wheel listener is implemented that way that scrolling is done in step of
	 * the height of the lines in the file list.
	 */
	private class ScrollList extends JScrollPane {
		final int yborder = 5;

		MouseWheelListener mouseWheelListener = new MouseWheelListener() {
			int yprev = 0;

			public void mouseWheelMoved(MouseWheelEvent e) {
				JViewport v = ScrollList.this.getViewport();
				Point p = v.getViewPosition();
				FileList list = (FileList) v.getView();
				p.y += e.getWheelRotation() * list.lineheight;
				p.y = Math.max(0, p.y);
				p.y = Math.min(p.y, list.getHeight() - (ScrollList.this.getHeight() - yborder * 2));
				if (p.y != yprev) {
					v.setViewPosition(p);
					yprev = p.y;
				}
			}
		};

		public ScrollList() {
			super();
			setBackground(BACKGROUND);
			setBorder(new Borders(yborder, 0, yborder, 0));
			for (MouseWheelListener l : getMouseWheelListeners()) {
				removeMouseWheelListener(l);
			}
			addMouseWheelListener(mouseWheelListener);
		}

		public void init(File[] files) {
			setViewportView(new FileList(files));
		}
	}

	/**
	 * The filename on bottom of the list
	 */
	private class Footer extends JPanel {
		JTextField g_filename = null;

		public Footer() {
			super(new BorderLayout());
			setBackground(BACKGROUND);
			JLabel label = new JLabel("Filename");
			label.setPreferredSize(new Dimension(60, 20));
			add(label, BorderLayout.WEST);
			add(g_filename = new JTextField(), BorderLayout.CENTER);
		}

		public String getFilename() {
			return g_filename.getText().trim();
		}

		public void setFilename(String filename) {
			g_filename.setText(filename.trim());
		}
	}

	/**
	 * Create a file selector for the path. The path may be a directory or a filename.
	 * 
	 * @param path - the pathname
	 * @param hidden - false without hidden files, true with hidden files
	 */
	public FileSelector(String path, boolean hidden) {
		super(new BorderLayout());
		this.hidden = hidden;
		setBackground(BACKGROUND);
		add(g_header = new Header(), BorderLayout.NORTH);
		add(g_filelist = new ScrollList(), BorderLayout.CENTER);
		add(g_footer = new Footer(), BorderLayout.SOUTH);
		init(new File(path != null ? path : Utils.getHomeDirectory()));
	}

	/**
	 * Create a file selector for the path. The path may be a directory or a filename. the selected files must not be
	 * hidden!
	 * 
	 * @param path - the pathname
	 */
	public FileSelector(String path) {
		this(path, false);
	}

	/**
	 * Returns the selected directory.
	 */
	public String getDirectory() {
		if (selectedPath.isDirectory()) {
			return selectedPath.getAbsolutePath();
		} else {
			return selectedPath.getParentFile().getAbsolutePath();
		}
	}

	/**
	 * Returns the selected, full specified filename.
	 */
	public String getFilename() {
		if (selectedPath.isFile()) {
			return selectedPath.getParent() + File.separator + g_footer.getFilename();
		} else {
			return selectedPath + File.separator + g_footer.getFilename();
		}
	}

	public void setDirectory(String path) {
		init(new File(path));
	}

	public void setFilename(String path) {
		init(new File(path));
	}

	// TODO: On loading file: the file MUST exist
	public boolean isOk() {
		return true;
	}

	public void reset() {
		init(savedPath);
	}

	public void saveState() {
		savedPath = selectedPath;
	}

	protected void enableFilename(boolean state) {
		g_footer.setVisible(state);
	}

	// **
	// Private
	// **

	private void setSelectedFile(File file) {
		selectedPath = file;
		g_footer.setFilename(selectedPath.getName());
	}

	private void init(File f) {
		if (savedPath == null) {
			savedPath = f;
		}
		selectedPath = f;
		if (f.isFile()) {
			setSelectedFile(f);
			init(f.getParentFile());
		} else if (f.isDirectory()) {
			File[] files = f.listFiles(new FileFilter() {
				@Override
				public boolean accept(File pathname) {
					return !pathname.isHidden() || hidden;
				}
			});
			if (files != null) {
				Arrays.sort(files, new Comparator<File>() {
					public int compare(File o1, File o2) {
						return o1.getAbsolutePath().compareTo(o2.getAbsolutePath());
					}
				});
				g_header.init(f);
				g_filelist.init(files);
				FileSelector.this.validate();
				FileSelector.this.repaint();
			}
		}
	}
}
