package de.bioinf.ui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JPanel;

/**
 * A dialog for typing or selecting input data. The dialog has a title, an input area and ok and cancel buttons.
 * Selecting the ok or cancel button disables the dialog. Pushing the ok button yields to a initial check of the input
 * data. This checking is send to the single input components.
 * 
 * <p>
 * The parameter T is the input component with the input fields etc. Using the method getInputComponent() the access to
 * this input area is provided.
 * 
 * @author ng
 */
public abstract class InputDialog<T extends JComponent> extends JDialog {
	private T g_input = null;

	class OkButton extends Button implements ActionListener {
		public OkButton() {
			super("Ok");
			addActionListener(this);
		}

		public void actionPerformed(ActionEvent e) {
			if (isOk()) {
				InputDialog.this.setVisible(false);
				saveState();
				afterOk(e);
			}
		}
	}

	class CancelButton extends Button implements ActionListener {
		public CancelButton() {
			super("Cancel");
			addActionListener(this);
		}

		public void actionPerformed(ActionEvent e) {
			InputDialog.this.setVisible(false);
			InputDialog.this.reset();
		}
	}

	/**
	 * Creates a dialog of width and height with the title.
	 * 
	 * @param comp - The component with the input fields.
	 */
	public InputDialog(String title, T comp, int width, int height) {
		super();

		setTitle(title);
		setSize(width, height);
		setLayout(new BorderLayout());

		add(BorderLayout.CENTER, g_input = comp);
		JPanel buttons = new JPanel(new GridLayout(1, 2, 2, 0));
		buttons.setBorder(new Borders(2, 1, 1, 1));
		buttons.add(new OkButton());
		buttons.add(new CancelButton());
		add(BorderLayout.SOUTH, buttons);
	}

	/**
	 * Implement this method to act on selecting the OK button and initial validation of the input fields.
	 * 
	 * @see isOk()
	 */
	abstract public void afterOk(ActionEvent e);

	/**
	 * Returns the component with the input fields.
	 */
	public T getInputComponent() {
		return g_input;
	}

	/**
	 * Check if the input fields are valid.
	 * 
	 * @see afterOk()
	 */
	public boolean isOk() {
		return isOk(g_input);
	}

	/**
	 * Reset to the previous state (either at initialization or after clicking on ok).
	 */
	public void reset() {
		reset(g_input);
	}

	/**
	 * Save the state to reset the component (after cancel click).
	 */
	public void saveState() {
		saveState(g_input);
	}

	// **
	// Private
	// **

	private boolean isOk(Component comp) {
		if (comp instanceof InputComponent) {
			if (!((InputComponent) comp).isOk()) {
				return false;
			}
		}
		if (comp instanceof JComponent) {
			for (Component c : ((JComponent) comp).getComponents()) {
				if (!isOk(c)) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Reset all subcomponents
	 */
	private void reset(Component comp) {
		if (comp instanceof InputComponent) {
			((InputComponent) comp).reset();
		}
		if (comp instanceof JComponent) {
			for (Component c : ((JComponent) comp).getComponents()) {
				reset(c);
			}
		}
	}

	/**
	 * SAve sate of all subcomponents
	 */
	private void saveState(Component comp) {
		if (comp instanceof InputComponent) {
			((InputComponent) comp).saveState();
		}
		if (comp instanceof JComponent) {
			for (Component c : ((JComponent) comp).getComponents()) {
				saveState(c);
			}
		}
	}
}
