package de.bioinf.ui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Point;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.util.ArrayList;

import javax.swing.BoundedRangeModel;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JViewport;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TableColumnModelEvent;
import javax.swing.event.TableColumnModelListener;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableModel;

/**
 * A special view. The view is made of several components:
 * 
 * <p>
 * <code>+-----+----+--------------------------------------+-------+---+</code><br>
 * <code>|     | HH | HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH | HHHHH |   |</code><br>
 * <code>+-----+----+--------------------------------------+-------+---+</code><br>
 * <code>| HHH | VV | VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV | VVVVV | S |</code><br>
 * <code>| HHH | VV | VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV | VVVVV | S |</code><br>
 * <code>| HHH | VV | VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV | VVVVV | S |</code><br>
 * <code>+-----+----+--------------------------------------+-------+---+</code><br>
 * <code>|     |    | SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS |       |   |</code><br>
 * <code>+-----+-------------------------------------------+-------+---+</code><br>
 * 
 * <p>
 * where H are the headers, V are the value areas and S are the scrollbars. The middle value table
 * is horizontally scrollable, the other two not (they are fixed). Depending on the constructor
 * arguments the left and/or right value area are created or not.
 * 
 * @author ng
 */
public class Sheet extends JPanel {
	// The base model for the data.
	private TableModel model = null;

	private int rows = 0;
	private int cols1 = 0;
	private int cols2 = 0;
	private int cols3 = 0;

	// The elements of this view: headers, values areas and scrollbars
	private Header1 g_header1;
	private Header2 g_header2;
	private Header3 g_header3;
	private HeaderV g_headerv;
	private Value1 g_values1;
	private Value2 g_values2;
	private Value3 g_values3;
	private ScrollH g_scrollh;
	private ScrollV g_scrollv;

	// The height of all rows
	private int rowHeight = 20;
	// All scrollable tables that are created so far.
	private ArrayList<ScrollTable> tables = new ArrayList<ScrollTable>();
	// A switch used in selecting columns and rows.
	private boolean selectable = true;

	/**
	 * A wrapper model to get a column range of a table model.
	 */
	class TableSubModel implements TableModel {
		// The data
		private TableModel model = null;
		// The starting column
		private int colstart = 0;
		// The ending column
		private int colend = 0;

		public TableSubModel(TableModel model, int colstart, int colend) {
			super();
			this.model = model;
			this.colstart = colstart;
			this.colend = colend;
		}

		public void addTableModelListener(TableModelListener l) {
		}

		public Class<?> getColumnClass(int colindex) {
			return String.class;
		}

		public int getColumnCount() {
			return colend - colstart + 1;
		}

		public String getColumnName(int colindex) {
			return model.getColumnName(getIndex(colindex));
		}

		public int getRowCount() {
			return model.getRowCount();
		}

		public Object getValueAt(int rowindex, int colindex) {
			return model.getValueAt(rowindex, getIndex(colindex));
		}

		public boolean isCellEditable(int rowindex, int colindex) {
			return true;
		}

		public void removeTableModelListener(TableModelListener l) {
		}

		public void setValueAt(Object value, int rowindex, int colindex) {
			model.setValueAt(value, rowindex, getIndex(colindex));
		}

		private int getIndex(int colindex) {
			return colstart + colindex;
		}
	}

	/**
	 * A Renderer for the values. Values with 0 have a background white color.
	 */
	class ValueRenderer extends DefaultTableCellRenderer {
		public Component getTableCellRendererComponent(JTable table, Object value, boolean selected, boolean focussed,
				int row, int column) {
			if (value != null && value instanceof String) {
				Component renderer = super.getTableCellRendererComponent(table, value, selected, focussed, row, column);
				Object o = table.getValueAt(row, column);
				if (!"0".equals((String) o)) {
					renderer.setForeground(Color.BLACK);
				} else {
					renderer.setForeground(Color.WHITE);
				}
				return renderer;
			}
			return this;
		}
	}

	/**
	 * The graphical element for a scrollable table.
	 */
	abstract class ScrollTable extends JScrollPane implements TableColumnModelListener, ListSelectionListener,
			ChangeListener {
		JTable g_table = null;

		public ScrollTable(TableModel model) {
			super();
			setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
			setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);
			setViewportView(g_table = new JTable(model));
			getViewport().addChangeListener(this);
			g_table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
			g_table.setCellSelectionEnabled(true);
			g_table.setRowHeight(rowHeight);
			g_table.getColumnModel().addColumnModelListener(this);
			g_table.getSelectionModel().addListSelectionListener(this);
			tables.add(this);
		}

		// Select row/column. Row and column are TableSubModel specific!
		public void select(int row, int col) {
			g_table.setRowSelectionInterval(row, row);
			g_table.setColumnSelectionInterval(col, col);
		}

		public void columnAdded(TableColumnModelEvent e) {
		}

		public void columnMarginChanged(ChangeEvent e) {
		}

		public void columnMoved(TableColumnModelEvent e) {
		}

		public void columnRemoved(TableColumnModelEvent e) {
		}

		// Act on column selection
		public void columnSelectionChanged(ListSelectionEvent e) {
			g_table.valueChanged(e);
			if (!e.getValueIsAdjusting()) {
				if (selectable) {
					selectable = false;
					checkSelection();
					selectable = true;
				}
			}
		}

		// Act on row selection
		public void valueChanged(ListSelectionEvent e) {
			g_table.valueChanged(e);
			if (!e.getValueIsAdjusting()) {
				if (selectable) {
					selectable = false;
					checkSelection();
					selectable = true;
				}
			}
		}

		// Act on vieport changes
		public void stateChanged(ChangeEvent e) {
			checkVieport();
		}

		// Select and unselect cells.
		abstract protected void checkSelection();

		// Set the right values for viewport changes
		abstract protected void checkVieport();

		// Set the viewport of the scrollbars to the same values as the two
		// tables. This implicit sets the vieports in the other tables.
		protected void setViewport(ScrollTable ysource, ScrollTable xsource) {
			g_scrollv.setValue(ysource.getViewport().getViewPosition().y);
			if (xsource == g_values2) {
				g_scrollh.setValue(xsource.getViewport().getViewPosition().x);
			}
		}

		// Select row/column taken from source tables
		protected void setSelection(ScrollTable rsource, ScrollTable csource, ScrollTable rtarget, ScrollTable ctarget) {
			int row = rsource.g_table.getSelectedRow();
			int col = csource.g_table.getSelectedColumn();
			rtarget.g_table.setRowSelectionInterval(row, row);
			ctarget.g_table.setColumnSelectionInterval(col, col);
			if (rtarget != ctarget) {
				if (rtarget.g_table.getSelectedColumn() == -1) {
					rtarget.g_table.setColumnSelectionInterval(0, 0);
				}
				if (ctarget.g_table.getSelectedRow() == -1) {
					ctarget.g_table.setRowSelectionInterval(0, 0);
				}
			}
			// Every other table should be unselected
			for (int i = 0; i < tables.size(); i++) {
				ScrollTable t = tables.get(i);
				if (t != rsource && t != csource && t != rtarget && t != ctarget) {
					t.g_table.clearSelection();
				}
			}
		}
	}

	/**
	 * A horizontal header
	 */
	abstract class HeaderH_ extends ScrollTable {
		public HeaderH_(int cols) {
			super(new Table(1, cols));
			setMinimumSize(new Dimension(20, rowHeight));
			g_table.setBackground(new Color(230, 230, 230));
		}

		public void set(String value, int col) {
			g_table.setValueAt(value, 0, col);
		}

		public String get(int col) {
			return (String) g_table.getValueAt(0, col);
		}
	}

	/**
	 * A vertical header
	 */
	abstract class HeaderV_ extends ScrollTable {
		public HeaderV_(int rows) {
			super(new Table(rows, 1));
			setMinimumSize(new Dimension(50, 20));
			g_table.setBackground(new Color(230, 230, 230));
		}

		public void set(String value, int row) {
			g_table.setValueAt(value, row, 0);
		}

		public String get(int row) {
			return (String) g_table.getValueAt(row, 0);
		}
	}

	/**
	 * A value area (table)
	 */
	abstract class ValueArea_ extends ScrollTable {
		public ValueArea_(TableModel model) {
			super(model);
			g_table.setBackground(Color.WHITE);
			g_table.setDefaultRenderer(String.class, new ValueRenderer());
		}

		public void setValue(String value, int row, int col) {
			g_table.setValueAt(value, row, col);
		}

		public String getValue(int row, int col) {
			return (String) g_table.getValueAt(row, col);
		}

		public int getMaxXOffset() {
			return (int) (g_table.getSize().getWidth() - this.getSize().getWidth());
		}

		public int getMaxYOffset() {
			return (int) (g_table.getSize().getHeight() - this.getSize().getHeight());
		}
	}

	/**
	 * The horizontal scrollbar.
	 */
	class ScrollH extends JScrollBar {
		public ScrollH(int cols) {
			super(JScrollBar.HORIZONTAL, 0, 50, 0, 100);
		}
	}

	/**
	 * The vertical scrollbar.
	 */
	class ScrollV extends JScrollBar {
		public ScrollV(int rows) {
			super(JScrollBar.VERTICAL, 0, rowHeight, 0, 100);
		}
	}

	class Header1 extends HeaderH_ {
		public Header1(int cols) {
			super(cols);
		}

		public void checkSelection() {
			setSelection(g_headerv, g_header1, g_values1, g_values1);
		}

		public void checkVieport() {
			setViewport(g_headerv, g_header1);
		}
	}

	class Header2 extends HeaderH_ {
		public Header2(int cols) {
			super(cols);
		}

		public void checkSelection() {
			setSelection(g_headerv, g_header2, g_values2, g_values2);
		}

		public void checkVieport() {
			setViewport(g_headerv, g_header2);
		}
	}

	class Header3 extends HeaderH_ {
		public Header3(int cols) {
			super(cols);
		}

		public void checkSelection() {
			setSelection(g_headerv, g_header3, g_values3, g_values3);
		}

		public void checkVieport() {
			setViewport(g_headerv, g_header3);
		}
	}

	class HeaderV extends HeaderV_ {
		public HeaderV(int cols) {
			super(cols);
		}

		public void checkSelection() {
			if (g_header1 != null && g_header1.g_table.getSelectedColumn() != -1) {
				setSelection(g_headerv, g_header1, g_values1, g_values1);
			}
			if (g_header2.g_table.getSelectedColumn() != -1) {
				setSelection(g_headerv, g_header2, g_values2, g_values2);
			}
			if (g_header3 != null && g_header3.g_table.getSelectedColumn() != -1) {
				setSelection(g_headerv, g_header3, g_values3, g_values3);
			}
		}

		public void checkVieport() {
			if (g_header1 != null && g_header1.g_table.getSelectedColumn() != -1) {
				setViewport(g_headerv, g_header1);
			}
			if (g_header2.g_table.getSelectedColumn() != -1) {
				setViewport(g_headerv, g_header2);
			}
			if (g_header3 != null && g_header3.g_table.getSelectedColumn() != -1) {
				setViewport(g_headerv, g_header3);
			}
		}
	}

	class Value1 extends ValueArea_ {
		public Value1(TableModel m) {
			super(m);
		}

		public void checkSelection() {
			setSelection(g_values1, g_values1, g_headerv, g_header1);
		}

		public void checkVieport() {
			setViewport(g_values1, g_values1);
		}
	}

	class Value2 extends ValueArea_ {
		public Value2(TableModel m) {
			super(m);
		}

		public void checkSelection() {
			setSelection(g_values2, g_values2, g_headerv, g_header2);
		}

		public void checkVieport() {
			setViewport(g_values2, g_values2);
		}
	}

	class Value3 extends ValueArea_ {
		public Value3(TableModel m) {
			super(m);
		}

		public void checkSelection() {
			setSelection(g_values3, g_values3, g_headerv, g_header3);
		}

		public void checkVieport() {
			setViewport(g_values3, g_values3);
		}
	}

	/**
	 * A listener that acts on value changes of a scrollbar.
	 */
	class ScrollbarListener implements AdjustmentListener {
		static final int XOFFSET = 1;
		static final int YOFFSET = 2;

		int type = 0;
		int prevxoffset = 0;
		int prevyoffset = 0;

		public ScrollbarListener(int type) {
			this.type = type;
		}

		public void adjustmentValueChanged(AdjustmentEvent e) {
			if (e.getAdjustmentType() == AdjustmentEvent.TRACK) {
				int offset = Math.max(0, e.getValue());
				if (type == YOFFSET) {
					if (offset != prevyoffset) {
						if (g_values1 != null) {
							setYPos(g_values1.getViewport(), offset);
						}
						setYPos(g_values2.getViewport(), offset);
						if (g_values3 != null) {
							setYPos(g_values3.getViewport(), offset);
						}
						setYPos(g_headerv.getViewport(), offset);
						prevyoffset = offset;
					}
				} else {
					if (offset != prevxoffset) {
						setXPos(g_values2.getViewport(), offset);
						setXPos(g_header2.getViewport(), offset);
						prevxoffset = offset;
					}
				}
			}
		}
	}

	/**
	 * A listener for resetting the scrollbar maximum.
	 */
	class SheetListener implements ComponentListener {
		public void componentHidden(ComponentEvent e) {
		}

		public void componentMoved(ComponentEvent e) {
		}

		public void componentResized(ComponentEvent e) {
			BoundedRangeModel m = g_scrollv.getModel();
			m.setRangeProperties(m.getValue(), rowHeight, 0, g_values2.getMaxYOffset() + rowHeight, false);
			m = g_scrollh.getModel();
			m.setRangeProperties(m.getValue(), 100, 0, g_values2.getMaxXOffset() + 100, false);
		}

		public void componentShown(ComponentEvent e) {
		}
	}

	/**
	 * Act on wheel scrolling. Note: Calculating the offset for scrolling must be in limits
	 * (0..maxYOffset())
	 */
	class WheelListener implements MouseWheelListener {
		public void mouseWheelMoved(MouseWheelEvent e) {
			int offset = g_values2.getViewport().getViewPosition().y + e.getWheelRotation() * rowHeight;
			offset = Math.min(offset, g_values2.getMaxYOffset());
			offset = Math.max(offset, 0);
			if (g_values1 != null) {
				setYPos(g_values1.getViewport(), offset);
			}
			setYPos(g_values2.getViewport(), offset);
			if (g_values3 != null) {
				setYPos(g_values3.getViewport(), offset);
			}
			setYPos(g_headerv.getViewport(), offset);
		}
	}

	/**
	 * Creates a sheet with the number of rows and columns. The number of the fixed columns may be
	 * 0.
	 * 
	 * @param rows - Number of rows
	 * @param cols1 - Number of columns in the left part (fixed columns)
	 * @param cols2 - Number of columns in the middle part (movable columns)
	 * @param cols3 - Number of columns in the right part (fixed columns)
	 */
	public Sheet(int rows, int cols1, int cols2, int cols3) {
		this(new Table(rows, cols1 + cols2 + cols3), cols1, cols2, cols3);
	}

	/**
	 * Creates a sheet from the specified model. The number of the fixed columns may be 0.
	 * 
	 * @param model - The model with the data.
	 * @param cols1 - Number of columns in the left part (fixed columns)
	 * @param cols2 - Number of columns in the middle part (movable columns)
	 * @param cols3 - Number of columns in the right part (fixed columns)
	 */
	public Sheet(TableModel model, int cols1, int cols2, int cols3) {
		super(new GridBagLayout());

		this.model = model;
		this.rows = model.getRowCount();
		this.cols1 = cols1;
		this.cols2 = cols2;
		this.cols3 = cols3;

		int gridx = 0;
		add(g_headerv = new HeaderV(rows), GridBagConstraints.VERTICAL, gridx, 1, 0, 0);
		g_headerv.setMinimumSize(g_headerv.g_table.getPreferredSize());

		if (cols1 > 0) {
			gridx++;
			add(g_header1 = new Header1(cols1), GridBagConstraints.HORIZONTAL, gridx, 0, 0, 0);
			g_header1.setMinimumSize(g_header1.g_table.getPreferredSize());
			TableModel m = new TableSubModel(model, cols1, cols1 + cols2 - 1);
			add(g_values1 = new Value1(m), GridBagConstraints.BOTH, gridx, 1, 0, 0);
		}

		{
			gridx++;
			add(g_header2 = new Header2(cols2), GridBagConstraints.HORIZONTAL, gridx, 0, 0, 0);
			TableModel m = new TableSubModel(model, cols1, cols1 + cols2 - 1);
			add(g_values2 = new Value2(m), GridBagConstraints.BOTH, gridx, 1, 1, 1);
			add(g_scrollh = new ScrollH(cols2), GridBagConstraints.HORIZONTAL, gridx, 2, 0, 0);
		}

		if (cols3 > 0) {
			gridx++;
			add(g_header3 = new Header3(cols3), GridBagConstraints.HORIZONTAL, gridx, 0, 0, 0);
			g_header3.setMinimumSize(g_header3.g_table.getPreferredSize());
			TableModel m = new TableSubModel(model, cols1 + cols2, cols1 + cols2 + cols3 - 1);
			add(g_values3 = new Value3(m), GridBagConstraints.BOTH, gridx, 1, 0, 0);
		}

		gridx++;
		add(g_scrollv = new ScrollV(rows), GridBagConstraints.VERTICAL, gridx, 1, 0, 0);

		g_scrollh.setMaximum(g_values2.getMaxXOffset());
		g_scrollv.setMaximum(g_values2.getMaxYOffset());

		g_scrollh.addAdjustmentListener(new ScrollbarListener(ScrollbarListener.XOFFSET));
		g_scrollv.addAdjustmentListener(new ScrollbarListener(ScrollbarListener.YOFFSET));

		// Select the row/column (0,0)
		selectable = false;
		g_headerv.select(0, 0);
		if (cols1 > 0) {
			g_header1.select(0, 0);
			g_values1.select(0, 0);
		} else {
			g_header2.select(0, 0);
			g_values2.select(0, 0);
		}
		selectable = true;

		addComponentListener(new SheetListener());

		WheelListener listener = new WheelListener();
		if (g_values1 != null) {
			g_values1.addMouseWheelListener(listener);
		}
		g_values2.addMouseWheelListener(listener);
		if (g_values3 != null) {
			g_values3.addMouseWheelListener(listener);
		}
	}

	/**
	 * Returns the model
	 */
	public TableModel getModel() {
		return model;
	}

	/**
	 * Returns the number of all columns.
	 */
	public int getColumnCount() {
		return cols1 + cols2 + cols3;
	}

	/**
	 * Returns the number of rows.
	 */
	public int getRowCount() {
		return rows;
	}

	/**
	 * Returns the horizontal header at column. It returns null, if not available!
	 */
	public String getHHeader(int col) {
		if (col < cols1) {
			return g_header1.get(col);
		} else if (col < cols1 + cols2) {
			return g_header2.get(col - cols1);
		} else if (col < cols1 + cols2 + cols3) {
			return g_header3.get(col - (cols1 + cols2));
		}
		return null;
	}

	/**
	 * Returns the vertical header at row. It returns null, if not available!
	 */
	public String getVHeader(int row) {
		if (row < rows) {
			return g_headerv.get(row);
		}
		return null;
	}

	/**
	 * Returns the value at row, column. It returns null, if not available!
	 */
	public String getValue(int row, int col) {
		if (col < cols1) {
			return g_values1.getValue(row, col);
		} else if (col < cols1 + cols2) {
			return g_values2.getValue(row, col - cols1);
		} else if (col < cols1 + cols2 + cols3) {
			return g_values3.getValue(row, col - (cols1 + cols2));
		}
		return null;
	}

	/**
	 * Sets the width of a column
	 */
	public void setColumnWidth(int col, int width) {
		if (col < cols1) {
			g_header1.g_table.getColumnModel().getColumn(col).setPreferredWidth(width);
			g_values1.g_table.getColumnModel().getColumn(col).setPreferredWidth(width);
		} else if (col < cols1 + cols2) {
			g_header2.g_table.getColumnModel().getColumn(col - cols1).setPreferredWidth(width);
			g_values2.g_table.getColumnModel().getColumn(col - cols1).setPreferredWidth(width);
		} else if (col < cols1 + cols2 + cols3) {
			g_header3.g_table.getColumnModel().getColumn(col - (cols1 + cols2)).setPreferredWidth(width);
			g_values3.g_table.getColumnModel().getColumn(col - (cols1 + cols2)).setPreferredWidth(width);
		}
	}

	/**
	 * Sets the width of the vertical header
	 */
	public void setVHeaderWith(int width) {
		g_headerv.g_table.getColumnModel().getColumn(0).setPreferredWidth(width);
	}

	/**
	 * Sets the value at position (The column is the one spreaded over cols1, cols2 and cols3 in the
	 * constructor)
	 */
	public void setValue(String value, int row, int col) {
		if (col < cols1) {
			g_values1.setValue(value, row, col);
		} else if (col < cols1 + cols2) {
			g_values2.setValue(value, row, col - cols1);
		} else if (col < cols1 + cols2 + cols3) {
			g_values3.setValue(value, row, col - (cols1 + cols2));
		}
	}

	/**
	 * Set the horizontal header title.
	 */
	public void setHHeader(String value, int col) {
		if (col < cols1) {
			g_header1.set(value, col);
		} else if (col < cols1 + cols2) {
			g_header2.set(value, col - cols1);
		} else if (col < cols1 + cols2 + cols3) {
			g_header3.set(value, col - (cols1 + cols2));
		}
	}

	/**
	 * Set the vertical header title.
	 */
	public void setVHeader(String value, int row) {
		g_headerv.set(value, row);
	}

	/**
	 * Sets a header for column AND row!
	 */
	public void setHeader(String value, int row) {
		setHHeader(value, row);
		setVHeader(value, row);
	}

	// **
	// Private
	// **

	private void setXPos(JViewport v, int pos) {
		Point p = v.getViewPosition();
		p.x = pos;
		v.setViewPosition(p);
	}

	private void setYPos(JViewport v, int pos) {
		Point p = v.getViewPosition();
		p.y = pos;
		v.setViewPosition(p);
	}

	private void add(JComponent comp, int fill, int gridx, int gridy, int weightx, int weighty) {
		GridBagConstraints c = new GridBagConstraints();
		c.fill = fill;
		c.gridx = gridx;
		c.gridy = gridy;
		c.weightx = weightx;
		c.weighty = weighty;
		add(comp, c);
	}
}
