package de.bioinf.ui;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Shape;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.UIManager;
import javax.swing.table.TableModel;

import de.bioinf.utils.BioinfException;

/**
 * A special view with static and dynamic parts. Dynamic means there are columns which may be horizontally moved and
 * (static) columns on the left right side which are allways displayed. The view is made of several components:
 * 
 * <pre>
 * +----+----+------------------+----+---+
 * |    | HH | HHHHHHHHHHHHHHHH | HH |   |
 * +----+----+------------------+----+---+
 * | HH | VV | VVVVVVVVVVVVVVVV | VV | S |
 * | HH | VV | VVVVVVVVVVVVVVVV | VV | S |
 * | HH | VV | VVVVVVVVVVVVVVVV | VV | S |
 * +----+----+------------------+----+---+
 * |    |    | SSSSSSSSSSSSSSSS |    |   |
 * +----+-----------------------+----+---+
 * </pre>
 * 
 * <p>
 * where H are the headers, V are the value areas and S are the scrollbars. The middle value table is horizontally
 * scrollable, the other two not (they are fixed). Depending on the constructor arguments the left and/or right value
 * area are created or not.
 * </p>
 * 
 * <p>
 * Note: Display is done via Graphics.drawString(), and the element which will be printed is checked whether it must be
 * printed. That means: elements outside the viewable area are NOT drawn, to speed up printing!
 * </p>
 * 
 * @author ng
 */
public class TextSheet extends JPanel {
	// The underlaying data
	private TableModel model = null;
	// The number of static columns on the left side.
	private int leftCols = 0;
	// the number of static columns on the right side.
	private int rightCols = 0;
	// Indicator for auto layout
	private boolean autolayout = false;
	// Indicator for resize
	private boolean resize = false;
	// Indicator if the sheet is ready to display data.
	private boolean initialized = false;;

	private int rowHeaderWidth = 80;
	private int[] colWidths = null;
	private String[] colHeaders = null;
	private String[] rowHeaders = null;
	// The height of the rows - it depends on the font.
	private int rowHeight = 0;
	// The gap between the elements and panels
	private int gap = 1;
	// The gap of the value to the left and right bounds.
	private int xgap = 2;
	// Offsets for realizing scrolling
	private int xoffset = 0;
	private int yoffset = 0;

	private Color headerColor = new Color(0xe8, 0xe8, 0xe8);
	private Color valueColorOdd = new Color(0xf0, 0xf0, 0xf0);
	private Color valueColorEven = new Color(0xf8, 0xf8, 0xf8);
	private int scrollHeight = UIManager.getInt("ScrollBar.height");
	private int scrollWidth = UIManager.getInt("ScrollBar.width");

	// Graphical elements
	private RowHeader g_rowHeader = null;
	private ColHeader g_colHeader1 = null;
	private ColHeader g_colHeader2 = null;
	private ColHeader g_colHeader3 = null;
	private Matrix g_matrix1 = null;
	private Matrix g_matrix2 = null;
	private Matrix g_matrix3 = null;
	private ScrollBar g_colScrollBar = null;
	private ScrollBar g_rowScrollBar = null;

	private Color borderColor = new Color(128, 128, 128);

	class SheetListener implements ComponentListener {
		@Override
		public void componentHidden(ComponentEvent e) {
		}

		@Override
		public void componentMoved(ComponentEvent e) {
		}

		@Override
		public void componentResized(ComponentEvent e) {
			resize = true;
			repaint();
		}

		@Override
		public void componentShown(ComponentEvent e) {
			resize = true;
			repaint();
		}
	}

	class SheetMouseListener implements MouseListener, MouseMotionListener, MouseWheelListener {
		int x = 0;
		int y = 0;

		@Override
		public void mouseWheelMoved(MouseWheelEvent ev) {
			if (initialized) {
				g_rowScrollBar.setValue(yoffset + ev.getWheelRotation() * (rowHeight + gap));
			}
		}

		@Override
		public void mouseDragged(MouseEvent ev) {
			if (initialized) {
				xoffset = Math.max(xoffset + x - ev.getX(), 0);
				yoffset = Math.max(yoffset + y - ev.getY(), 0);
				xoffset = Math.min(xoffset, g_colScrollBar.getMaximum() - g_colScrollBar.getVisibleAmount());
				yoffset = Math.min(yoffset, g_rowScrollBar.getMaximum() - g_rowScrollBar.getVisibleAmount());
				g_colScrollBar.setValueWithoutListener(xoffset);
				g_rowScrollBar.setValueWithoutListener(yoffset);
				x = ev.getX();
				y = ev.getY();
				repaint();
			}
		}

		@Override
		public void mouseMoved(MouseEvent ev) {
		}

		@Override
		public void mouseClicked(MouseEvent ev) {
		}

		@Override
		public void mouseEntered(MouseEvent ev) {
		}

		@Override
		public void mouseExited(MouseEvent ev) {
		}

		@Override
		public void mousePressed(MouseEvent ev) {
			if (initialized) {
				x = ev.getX();
				y = ev.getY();
			}
		}

		@Override
		public void mouseReleased(MouseEvent arg0) {
		}
	}

	class ScrollBar extends JScrollBar implements AdjustmentListener {
		static final int XOFFSET = 1;
		static final int YOFFSET = 2;
		int type = 0;
		boolean active = true;

		public ScrollBar(int orientation) {
			super(orientation);
			type = orientation == JScrollBar.HORIZONTAL ? XOFFSET : YOFFSET;
			addAdjustmentListener(this);
		}

		public void setValueWithoutListener(int value) {
			active = false;
			super.setValue(value);
		}

		public void setValue(int value) {
			active = true;
			super.setValue(value);
		}

		public void adjustmentValueChanged(AdjustmentEvent e) {
			if (active) {
				if (e.getAdjustmentType() == AdjustmentEvent.TRACK) {
					int offset = Math.max(0, e.getValue());
					if (type == YOFFSET) {
						if (offset != yoffset) {
							yoffset = offset;
							g_matrix1.repaint();
							g_matrix2.repaint();
							g_matrix3.repaint();
							g_rowHeader.repaint();
						}
					} else {
						if (offset != xoffset) {
							xoffset = offset;
							g_colHeader2.repaint();
							g_matrix2.repaint();
						}
					}
				}
			}
		}
	}

	abstract class SheetPart extends JPanel {
		public SheetPart() {
			super(null);
			setBackground(borderColor);
		}

		public void drawString(Graphics g, String s, int x, int y, int width, Color bg) {
			int w = getWidth();
			int h = getHeight();
			if (x + width >= 0 && x <= w && y + rowHeight >= 0 && y <= h) {
				Shape clip = g.getClip();
				int wclip = x >= 0 ? (x + width < w ? width : w - x) : (x + width <= w ? width : w);
				g.setClip(x > 0 ? x : 0, y > 0 ? y : 0, wclip, y + rowHeight <= h ? rowHeight : h - y);
				g.setColor(bg);
				g.fillRect(x, y, width, rowHeight);
				g.setColor(Color.BLACK);
				g.drawString(s, x + xgap, y + rowHeight - 3);
				g.setClip(clip);
			}
		}
	}

	class RowHeader extends SheetPart {
		public RowHeader() {
			super();
		}

		@Override
		public void paint(Graphics g) {
			super.paint(g);
			for (int i = yoffset / (rowHeight + gap), y = i * (rowHeight + gap); i < model.getRowCount()
					&& y - yoffset < getHeight(); i++) {
				drawString(g, asString(rowHeaders[i]), 0, y - yoffset, rowHeaderWidth, headerColor);
				y += rowHeight + gap;
			}
		}
	}

	class ColHeader extends SheetPart {
		int min = 0;
		int max = 0;
		boolean xscroll = false;

		public ColHeader(int min, int max, boolean xscroll) {
			super();
			this.min = min;
			this.max = max;
			this.xscroll = xscroll;
		}

		@Override
		public void paint(Graphics g) {
			super.paint(g);
			for (int i = min, x = xscroll ? -xoffset : 0; i < max; i++) {
				drawString(g, asString(colHeaders[i]), x, 0, colWidths[i], headerColor);
				x += colWidths[i] + gap;
			}
		}
	}

	class Matrix extends SheetPart {
		int min = 0;
		int max = 0;
		boolean xscroll = false;

		public Matrix(int min, int max, boolean xscroll) {
			super();
			this.min = min;
			this.max = max;
			this.xscroll = xscroll;
		}

		@Override
		public void paint(Graphics g) {
			super.paint(g);

			// Calculate the left index (min) and the left offset (xmin) to avoid checking later.
			int lindex = min;
			int xmin = xscroll ? -xoffset : 0;
			for (; lindex < max && xmin + colWidths[lindex] < 0; lindex++) {
				xmin += colWidths[lindex] + gap;
			}

			for (int i = yoffset / (rowHeight + gap), y = i * (rowHeight + gap); i < model.getRowCount()
					&& y - yoffset < getHeight(); i++) {
				Color c = (i % 2) == 1 ? valueColorOdd : valueColorEven;
				for (int k = lindex, x = xmin; k < max && x < getWidth(); k++) {
					drawString(g, asString(model.getValueAt(i, k)), x, y - yoffset, colWidths[k], c);
					x += colWidths[k] + gap;
				}
				y += rowHeight + gap;
			}
		}
	}

	/**
	 * Create an empty sheet.
	 */
	public TextSheet() {
		super(null);
		addComponentListener(new SheetListener());
		SheetMouseListener listener = new SheetMouseListener();
		addMouseListener(listener);
		addMouseMotionListener(listener);
		addMouseWheelListener(listener);
	}

	/**
	 * Create a sheet of the specified size.
	 * 
	 * @param rows - Number of rows
	 * @param leftcols - Number of columns in the left part (static columns)
	 * @param middlecols - Number of columns in the middle part (dynamic columns)
	 * @param rightcols - Number of columns in the right part (static columns)
	 */
	public TextSheet(int rows, int leftcols, int middlecols, int rightcols) throws BioinfException {
		this();
		init(new Table(rows, leftcols + middlecols + rightcols), leftcols, rightcols);
	}

	/**
	 * Do automatic layout. That means to recalculate the width of all components to the minimum size.
	 * 
	 * Note: Call a repaint() if changes should be displayed!
	 */
	public void autolayout() {
		autolayout = true;
	}

	/**
	 * Initialize the sheet with a model and the number of columns left and right which are not scrolled.
	 * 
	 * Note: The width of all columns is initialized and must be set explicitly afterwards if desired!
	 * 
	 * @param model - The underlaying data.
	 * @param leftcols - Number of columns in the left part (static columns)
	 * @param rightcols - Number of columns in the right part (static columns)
	 */
	public void init(TableModel model, int leftcols, int rightcols) throws BioinfException {
		BioinfException.fire(model != null, "Please specify a table!");
		BioinfException.fire(leftcols + rightcols <= model.getColumnCount(), "Left + right columns exceeds table!");

		removeAll();

		this.model = model;
		this.leftCols = leftcols;
		this.rightCols = rightcols;

		colHeaders = new String[model.getColumnCount()];
		rowHeaders = new String[model.getRowCount()];
		colWidths = new int[model.getColumnCount()];
		for (int i = 0; i < colWidths.length; i++) {
			colWidths[i] = rowHeaderWidth;
		}

		add(g_rowHeader = new RowHeader());
		add(g_colHeader1 = new ColHeader(0, leftCols, false));
		add(g_matrix1 = new Matrix(0, leftCols, false));
		add(g_colHeader2 = new ColHeader(leftCols, model.getColumnCount() - rightCols, true));
		add(g_matrix2 = new Matrix(leftCols, model.getColumnCount() - rightCols, true));
		add(g_colScrollBar = new ScrollBar(JScrollBar.HORIZONTAL));
		add(g_colHeader3 = new ColHeader(model.getColumnCount() - rightCols, model.getColumnCount(), false));
		add(g_matrix3 = new Matrix(model.getColumnCount() - rightCols, model.getColumnCount(), false));
		add(g_rowScrollBar = new ScrollBar(JScrollBar.VERTICAL));

		initialized = true;
	}

	public int getRowCount() {
		return model.getRowCount();
	}

	public int getColCount() {
		return model.getColumnCount();
	}

	public String getValue(int row, int col) {
		return (String) model.getValueAt(row, col);
	}

	public String getColumnHeader(int index) {
		return colHeaders[index];
	}

	public String getRowHeader(int index) {
		return rowHeaders[index];
	}

	public void setValue(String value, int row, int col) {
		model.setValueAt(value, row, col);
	}

	public void setRowHeader(String title, int index) {
		rowHeaders[index] = title;
	}

	public void setColumnHeader(String title, int index) {
		colHeaders[index] = title;
	}

	public void setColumnWidth(int width, int index) {
		colWidths[index] = width;
		resize = true;
	}

	public void setRowHeaderWidth(int width) {
		rowHeaderWidth = width;
		resize = true;
	}

	public void setHeaderColor(Color c) {
		headerColor = c;
	}

	public void setValueColorOdd(Color c) {
		valueColorOdd = c;
	}

	public void setValueColorEven(Color c) {
		valueColorEven = c;
	}

	/**
	 * Return the presentation of the header/value object. It may be overwritten.
	 */
	public String asString(Object o) {
		return o != null ? o.toString() : "";
	}

	@Override
	public void paint(Graphics g) {
		if (rowHeight == 0) {
			rowHeight = g.getFontMetrics().getHeight() + 2;
		}
		autolayout(g);
		resize();
		super.paint(g);
	}

	// **
	// Private
	// **

	/**
	 * Resizes the component so that the column width fits best.
	 */
	private void autolayout(Graphics g) {
		if (autolayout) {
			FontMetrics m = g.getFontMetrics();
			rowHeaderWidth = 0;
			for (String s : rowHeaders) {
				rowHeaderWidth = Math.max(rowHeaderWidth, m.stringWidth(asString(s)));
			}
			rowHeaderWidth += 2 * xgap;
			for (int k = 0; k < model.getColumnCount(); k++) {
				colWidths[k] = m.stringWidth(asString(colHeaders[k]));
				for (int i = 0; i < model.getRowCount(); i++) {
					colWidths[k] = Math.max(colWidths[k], m.stringWidth(asString(getValue(i, k))));
				}
				colWidths[k] += 2 * xgap;
			}
			autolayout = false;
			resize = true;
		}
	}

	/**
	 * In case of activation or resize do layout.
	 */
	private void resize() {
		if (resize && model != null) {
			Insets insets = getInsets();

			int leftWidth = sumWidths(0, leftCols);
			int rightWidth = sumWidths(model.getColumnCount() - rightCols, model.getColumnCount());
			int middleWidth = getWidth()
					- (rowHeaderWidth + leftWidth + rightWidth + scrollWidth + 4 * gap + insets.left + insets.right);
			int middleHeight = getHeight() - (rowHeight + scrollHeight + 2 * gap + insets.top + insets.bottom);
			int modelHeight = model.getRowCount() * rowHeight + (model.getRowCount() - 1) * gap;
			int modelWidth = sumWidths(leftCols, model.getColumnCount() - rightCols);

			middleHeight = Math.min(middleHeight, modelHeight);
			middleWidth = Math.min(middleWidth, modelWidth);

			int x = insets.left;
			g_rowHeader.setLocation(x, rowHeight + gap + insets.top);
			g_rowHeader.setSize(rowHeaderWidth, middleHeight);

			x += rowHeaderWidth + gap;
			g_colHeader1.setLocation(x, insets.top);
			g_colHeader1.setSize(leftWidth, rowHeight);
			g_matrix1.setLocation(x, rowHeight + gap + insets.top);
			g_matrix1.setSize(leftWidth, middleHeight);

			x += leftWidth + gap;
			g_colHeader2.setLocation(x, insets.top);
			g_colHeader2.setSize(middleWidth, rowHeight);
			g_matrix2.setLocation(x, rowHeight + gap + insets.top);
			g_matrix2.setSize(middleWidth, middleHeight);
			g_colScrollBar.setLocation(x, middleHeight + rowHeight + 2 * gap + insets.top);
			g_colScrollBar.setSize(middleWidth, scrollHeight);

			x += middleWidth + gap;
			g_colHeader3.setLocation(x, insets.top);
			g_colHeader3.setSize(rightWidth, rowHeight);
			g_matrix3.setLocation(x, rowHeight + gap + insets.top);
			g_matrix3.setSize(rightWidth, middleHeight);

			x += rightWidth + gap;
			g_rowScrollBar.setLocation(x, rowHeight + gap + insets.top);
			g_rowScrollBar.setSize(scrollWidth, middleHeight);

			g_colScrollBar.setVisibleAmount(middleWidth);
			g_colScrollBar.setMinimum(0);
			g_colScrollBar.setMaximum(modelWidth - middleWidth + g_colScrollBar.getVisibleAmount());
			g_rowScrollBar.setVisibleAmount(rowHeight + gap);
			g_rowScrollBar.setMinimum(0);
			g_rowScrollBar.setMaximum(modelHeight - middleHeight + g_rowScrollBar.getVisibleAmount());

			resize = false;
		}
	}

	/**
	 * Sum the with of the columns min to max.
	 */
	private int sumWidths(int min, int max) {
		int width = 0;
		for (int i = min; i < max; i++) {
			width += colWidths[i] + (i > min ? gap : 0);
		}
		return width;
	}
}
