package de.bioinf.ui;

import java.awt.Component;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;

import javax.swing.DefaultListModel;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JScrollPane;
import javax.swing.ListCellRenderer;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import de.bioinf.appl.Colors;
import de.bioinf.utils.SelectionEvent;
import de.bioinf.utils.SelectionListener;

/**
 * A simple class for displaying values in a list.
 * 
 * @author ng
 */
public class ValueList extends JScrollPane {
	private JList g_list = null;
	private DefaultListModel model = null;

	private ArrayList<SelectionListener<ValueList>> listeners = new ArrayList<SelectionListener<ValueList>>();

	private class ListListener implements ListSelectionListener {
		public void valueChanged(ListSelectionEvent e) {
			if (!e.getValueIsAdjusting()) {
				for (SelectionListener<ValueList> listener : listeners) {
					listener.selected(new SelectionEvent<ValueList>(ValueList.this));
				}
			}
		}
	}

	private class Renderer extends JLabel implements ListCellRenderer {
		public Component getListCellRendererComponent(JList list, Object value, int index, boolean selected,
				boolean focussed) {
			setText((String) value);
			setBackground(selected ? Colors.SELECTED_BG : list.getBackground());
			setForeground(list.getForeground());
			setEnabled(list.isEnabled());
			setOpaque(true);
			return this;
		}
	}

	public ValueList() {
		super();
		g_list = new JList(model = new DefaultListModel());
		g_list.addListSelectionListener(new ListListener());
		setRenderer(new Renderer());
		setViewportView(g_list);
	}

	public ValueList(String[] values) {
		this();
		init(values);
	}

	public void addSelectionListener(SelectionListener<ValueList> listener) {
		listeners.add(listener);
	}

	/**
	 * Add a value to the end of the list
	 */
	public void addValue(String value) {
		model.addElement(value);
	}

	/**
	 * Initialize with the values
	 */
	public void init(String[] values) {
		model.clear();
		if (values != null) {
			for (String value : values) {
				model.addElement(value);
			}
		}
	}

	/**
	 * Set the cell renderer. The default one used here may be overwritten.
	 */
	public void setRenderer(ListCellRenderer renderer) {
		g_list.setCellRenderer(renderer);
	}

	public void setVisibleItems(int count) {
		g_list.setVisibleRowCount(count);
	}

	/**
	 * Returns the selected elements as an array.
	 */
	public String[] getSelection() {
		int[] indices = g_list.getSelectedIndices();
		String[] selection = new String[indices.length];
		for (int i = 0; i < indices.length; i++) {
			selection[i] = (String) model.get(indices[i]);
		}
		return selection;
	}

	public boolean isSelected() {
		return !g_list.isSelectionEmpty();
	}

	/**
	 * Adds the values to the selected elements
	 */
	public void addSelection(String[] values) {
		HashSet<String> h = new HashSet<String>(Arrays.asList(values));
		for (int i = 0; i < model.size(); i++) {
			if (h.contains(model.getElementAt(i))) {
				g_list.addSelectionInterval(i, i);
			}
		}
	}

	/**
	 * Removes the values from the selected elements
	 */
	public void removeSelection(String[] values) {
		HashSet<String> h = new HashSet<String>(Arrays.asList(values));
		for (int i = 0; i < model.size(); i++) {
			if (h.contains(model.getElementAt(i))) {
				g_list.removeSelectionInterval(i, i);
			}
		}
	}

	/**
	 * Select the values and deselect all others
	 */
	public void select(String[] values) {
		selectNone();
		addSelection(values);
	}

	/**
	 * Select this value and deselect all others
	 */
	public void select(String value) {
		g_list.setSelectedValue(value, true);
	}

	public void selectAll() {
		g_list.setSelectionInterval(0, g_list.getModel().getSize() - 1);
	}

	public void selectNone() {
		g_list.clearSelection();
	}
}
