package de.bioinf.ui.graph;

import java.awt.Graphics;

import de.bioinf.utils.BioinfException;
import de.bioinf.utils.ValueRange;

/**
 * The base class for elements (in this case containing function values) to be added to a graph panel.
 * 
 * @author ng
 */
public class PolygonElem extends GraphElem {
	/**
	 * The base panel (upon which this element will be displayed). That panel is used to transform x/y values of the
	 * element to the corresponding x/y values of the display.
	 */
	private CoordPanel panel = null;
	/**
	 * The original x value
	 */
	private PolygonValuePair[] values = null;

	private static final double DISTANCE = 5.0;

	// The size of the displayed area
	private ValueRange xrangeElem = new ValueRange();
	private ValueRange yrangeElem = new ValueRange();

	/**
	 * Creates a graph element, but does NOT add it to the underlaying panel.
	 * 
	 * @param values - the value pairs for the graph
	 * @param id - information that are displayed when clicked.
	 */
	public PolygonElem(CoordPanel panel, PolygonValuePair[] values, String id) throws BioinfException {
		super(id);
		BioinfException.fire(values != null && values.length > 1, "No valid values for polygon " + id);
		this.panel = panel;
		this.values = values;
	}

	@Override
	public void adjustRanges(ValueRange xrange, ValueRange yrange) {
		for (PolygonValuePair pair : values) {
			xrange.add(pair.xval);
			yrange.add(pair.yval);
			xrangeElem.add(pair.xval);
			yrangeElem.add(pair.yval);
		}
	}

	/**
	 * Display a polygon element only if the occupied area is part of the underlaying graphics panel.
	 */
	@Override
	public boolean isVisible() {
		return panel.getX(xrangeElem.getMaxVal()) >= 0 && panel.getX(xrangeElem.getMinVal()) <= panel.getWidth()
				&& panel.getY(yrangeElem.getMinVal()) > 0 && panel.getY(yrangeElem.getMaxVal()) <= panel.getHeight();
	}

	/**
	 * Is the Element hit on position? This is evaluated by calculating the distance of the point (x,y) with the points
	 * of the polygon. If the distance of 2 polygon points is to far, points between this polygon points will be
	 * calculated and compared to the given point (x,y).
	 */
	@Override
	public boolean isHit(int x, int y) {
		for (int i = 0; i < values.length; i++) {
			int x1 = panel.getX(values[i].xval);
			int y1 = panel.getY(values[i].yval);
			if (distance(x, y, x1, y1) < DISTANCE) {
				return true;
			} else {
				if (i < values.length - 1) {
					int x2 = panel.getX(values[i + 1].xval);
					int y2 = panel.getY(values[i + 1].yval);
					double dist = distance(x1, y1, x2, y2);
					if (dist > DISTANCE) {
						double xdiff = (x2 - x1) / (dist / DISTANCE + 1.0);
						double ydiff = (y2 - y1) / (dist / DISTANCE + 1.0);
						double xd = x1;
						double yd = y1;
						while (xd + xdiff < x2) {
							if (distance(x, y, (int) xd, (int) yd) < DISTANCE) {
								return true;
							}
							xd += xdiff;
							yd += ydiff;
						}
					}
				}
			}
		}
		return false;
	}

	public void paint(Graphics g) {
		int[] xs = new int[values.length];
		int[] ys = new int[values.length];
		for (int i = 0; i < values.length; i++) {
			xs[i] = panel.getX(values[i].xval);
			ys[i] = panel.getY(values[i].yval);
		}
		g.setColor(getBackground());
		g.drawPolyline(xs, ys, xs.length);
	}

	protected CoordPanel getPanel() {
		return panel;
	}

	/**
	 * Returns the value pairs
	 */
	protected PolygonValuePair[] getValues() {
		return values;
	}

	// **
	// Private
	// **

	/**
	 * Returns the distance between 2 points.
	 */
	private double distance(int x1, int y1, int x2, int y2) {
		return Math.sqrt(Math.pow(x2 - x1, 2) + Math.pow(y2 - y1, 2));
	}
}
