package de.bioinf.utils;

import java.util.Hashtable;
import java.util.Stack;

/**
 * This is a simple calculator for evaluating terms. That means the +,-,*, and /
 * operators are available, as the usage of round braces. Calculating *,/ before
 * *,- is regarded as the usage of functions like: f(1,2). The name and meaning
 * of the functions depends on the underlaying subclass implementing them.
 */
public abstract class Calculator {
	// The arguments for the operations/functions
	private Stack<String> args = new Stack<String>();
	// The operations/functions
	private Stack<String> ops = new Stack<String>();
	// The position in the token array.
	private int tokindex = 0;

	// The precedence of the operators
	private static Hashtable<String, Integer> precs = new Hashtable<String, Integer>();

	public Calculator() {
	}

	/**
	 * Evaluate the term
	 */
	public synchronized double eval(String term) throws BioinfException {
		try {
			tokindex = 0;
			eval(Utils.toArray(term, "()*/+-,", true), 0);
			return Double.parseDouble(args.pop());
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage("Term " + term + " is not evaluable!");
		}
	}

	/**
	 * Is the token a function?
	 */
	public abstract boolean isOp(String token);

	/**
	 * Evaluate a special operation/function with the given arguments.
	 */
	public abstract double eval(String op, String[] args);

	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("Args=").append(args.toString());
		buf.append(", Ops=").append(ops.toString());
		return buf.toString();
	}

	/**
	 * Do the evaluation, extracting the tokens and filling stacks.
	 */
	private void eval(String[] tokens, int argpos) throws BioinfException {
		while (tokindex < tokens.length) {
			String token = tokens[tokindex++].trim();
			if ("".equals(token)) {
				// Avoid using blanks
			} else if ("(".equals(token)) {
				eval(tokens, args.size());
			} else if (")".equals(token)) {
				break;
			} else if (",".equals(token)) {
				// Evaluate missing operations +-/*
				while (ops.size() > 0 && "+-*/".indexOf(ops.lastElement()) >= 0) {
					eval(argpos);
				}
			} else if ("+-*/".indexOf(token) >= 0) {
				if (!isPrec(token)) {
					eval(argpos);
				}
				ops.push(token);
			} else {
				if (isOp(token)) {
					ops.push(token);
				} else {
					args.push(token);
				}
			}
		}
		// Evaluate all missing operations
		while (eval(argpos))
			;
	}

	/**
	 * Evaluate the arguments with the operator on the top of ops. Before
	 * evaluation the number of arguments is checked, because operations may be
	 * pre-, inside- or postspecifeid. That means: "+(1,2)" is the same as "1+2"
	 * or "(1,2)+", it works all the same!
	 */
	private boolean eval(int argpos) throws BioinfException {
		if (ops.size() == 0) {
			return false;
		}
		try {
			String op = ops.lastElement();
			// Binary operators
			if ("+-*/".indexOf(op) >= 0) {
				if (countArgs(argpos) >= 2) {
					double arg2 = Double.parseDouble(args.pop());
					double arg1 = Double.parseDouble(args.pop());
					switch (op.charAt(0)) {
					case '+':
						args.push(new Double(arg1 + arg2).toString());
						break;
					case '-':
						args.push(new Double(arg1 - arg2).toString());
						break;
					case '*':
						args.push(new Double(arg1 * arg2).toString());
						break;
					case '/':
						args.push(new Double(arg1 / arg2).toString());
						break;
					}
					ops.pop();
				} else {
					return false;
				}
			}
			// Other in subclasses defined operators.
			else {
				String[] opargs = new String[args.size() - argpos];
				for (int i = opargs.length - 1; i >= 0; i--) {
					opargs[i] = args.pop();
				}
				args.push(new Double(eval(op, opargs)).toString());
				ops.pop();
			}
		} catch (Exception e) {
			throw BioinfException.convert(e);
		}
		return true;
	}

	/**
	 * Count the arguments starting at stack position pos.
	 */
	private int countArgs(int argpos) {
		int count = 0;
		for (int i = argpos; i < args.size(); i++) {
			count++;
		}
		return count;
	}

	/**
	 * Has this operator precedence of the last on the stack?
	 */
	private boolean isPrec(String op) {
		if (ops.size() > 0) {
			int prec1 = precs.containsKey(op) ? precs.get(op).intValue() : 0;
			int prec2 = precs.containsKey(ops.lastElement()) ? precs.get(ops.lastElement()).intValue() : 0;
			return prec1 > prec2;
		}
		return true;
	}

	static {
		precs.put("+", new Integer(1));
		precs.put("-", new Integer(1));
		precs.put("*", new Integer(2));
		precs.put("/", new Integer(2));
	}
}
