package de.bioinf.utils;

import java.util.ArrayList;

/**
 * A class for reading and preparing configuration files. Comments start with a '#' sign and reach to the end of the
 * line. A section may be specified several times. In order of appearance the values may be overwritten while loading
 * the content. The data file has a structure like:
 * 
 * <pre>
 * [section-name-1]
 * key1 = value1
 * ...
 * [section-name-2]
 * key2 = value2
 * ...
 * </pre>
 * 
 * @see ConfigSection
 * @author ng
 */
public class ConfigFile {
	private String[] lines = null;

	/**
	 * Reads the source line by line and prepares for further processing.
	 */
	public ConfigFile(Source source) throws BioinfException {
		prepare(source);
	}

	/**
	 * Returns the lines that are inside a given section. The section itself is written as "[text]" in the file, but the
	 * section name as the parameter must not have brackets! The section may be spread along the file and may be
	 * specified multiple times.
	 */
	public String[] getSection(String name) {
		String s = "[" + name + "]";
		boolean ok = false;
		ArrayList<String> a = new ArrayList<String>();
		for (String line : lines) {
			if (!line.startsWith("[") && ok) {
				a.add(line);
			}
			if (line.equals(s)) {
				ok = true;
			} else if (line.startsWith("[")) {
				ok = false;
			}
		}
		return a.toArray(new String[a.size()]);
	}

	// **
	// Private
	// **

	private void prepare(Source source) throws BioinfException {
		try {
			LineReader reader = source.getReader(false);
			String line = null;
			ArrayList<String> a = new ArrayList<String>();
			while ((line = reader.readLine()) != null) {
				line = prepare(line);
				if (line.length() > 0) {
					a.add(line);
				}
			}
			reader.close();
			lines = a.toArray(new String[a.size()]);
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage(
					String.format("Cannot load configuration '%s'!", source.getLocation()));
		}
	}

	/**
	 * Remove comments and trim the line.
	 */
	private String prepare(String line) {
		int index = line.indexOf("#");
		if (index >= 0) {
			line = line.substring(0, index);
		}
		return line.trim();
	}
}
