package de.bioinf.utils;

/**
 * A simple array implementation for adding elements to an array that keeps track automatically about the internal
 * capacity. The array just increases its size.
 * 
 * @author ng
 */
public abstract class DynamicArray<T> {
	private T[] data = null;
	private int usedsize = 0;
	/**
	 * The incremental size used in the resize() call.
	 */
	private int incsize;

	public DynamicArray() {
		this(16);
	}

	/**
	 * Creates a dynamic array.
	 * 
	 * @param incsize - incremental size. The amount of enlargement, if the internal array must be resized.
	 */
	public DynamicArray(int incsize) {
		data = getArray(this.incsize = incsize);
	}

	/**
	 * Adds an element to the end of the array
	 */
	public void add(T value) {
		set(usedsize, value);
	}

	/**
	 * Returns the element at position index.
	 */
	public T get(int index) {
		return data[index];
	}

	/**
	 * This method must be implemented in a "real" class.
	 */
	public abstract T[] getArray(int size);

	/**
	 * Sets the element at position index. If the position index is beyond the actual size, the array will be extended.
	 */
	public void set(int index, T value) {
		enlarge(index);
		data[index] = value;
		usedsize = Math.max(usedsize, index + 1);
	}

	/**
	 * Returns the size used so far.
	 */
	public int size() {
		return usedsize;
	}

	/**
	 * Returns the elements as an array (as a copy).
	 */
	public T[] toArray() {
		T[] newdata = getArray(usedsize);
		System.arraycopy(data, 0, newdata, 0, usedsize);
		return newdata;
	}

	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("[");
		if (usedsize > 0) {
			buf.append(data[0]);
			for (int i = 1; i < usedsize; i++) {
				buf.append(",").append(data[i]);
			}
		}
		buf.append("]");
		return buf.toString();
	}

	// **
	// Private
	// **

	private void enlarge(int index) {
		if (data.length < index) {
			T[] olddata = data;
			data = getArray(index + incsize);
			System.arraycopy(olddata, 0, data, 0, olddata.length);
		}
	}
}
