package de.bioinf.utils;

import java.io.File;
import java.io.FilenameFilter;
import java.util.Arrays;
import java.util.Comparator;

/**
 * A class to get commands from files in a directory that match a given suffix.
 * Every file contains a single command that will be executed using
 * <code>perform(String)</code> in a subclass. Performed files will be
 * deleted, file that result in an error are renamed to filename.err. The files
 * are "executed" in order using the modification time of the file!
 * 
 * @author ng
 */
public abstract class FileCommandRunner {
	private File dir = null;
	// The suffix for command files
	private String suffix = null;

	/**
	 * Compare (and afterwards sort) the files by modification time.
	 */
	class FileComparator implements Comparator<File> {
		public int compare(File f1, File f2) {
			long m1 = f1.lastModified();
			long m2 = f2.lastModified();
			if (m1 > m2)
				return -1;
			else if (m1 == m2)
				return 0;
			else
				return 1;
		}
	}

	/**
	 * Filter files by name (*.cmd)
	 */
	class FileFilter implements FilenameFilter {
		public boolean accept(File dir, String name) {
			return name.endsWith(suffix);
		}
	}

	/**
	 * Creates a runner that will take the command files from the given
	 * directory.
	 */
	public FileCommandRunner(String dirname, String suffix) throws BioinfException {
		try {
			dir = new File(dirname);
			if (!dir.exists()) {
				throw new BioinfException(String.format("Directory %s does not exist!", dir));
			}
			if (!dir.isDirectory()) {
				throw new BioinfException(String.format("%s is not a directory!", dir));
			}
			this.suffix = suffix;
		} catch (Exception e) {
			throw BioinfException.convert(e);
		}
	}

	/**
	 * Start the process in an infinite loop.
	 */
	public void run() {
		try {
			while (true) {
				File[] files = dir.listFiles(new FileFilter());
				Arrays.sort(files, new FileComparator());
				for (File f : files) {
					String line = null;
					try {
						LineReader reader = new LineReader(f.getAbsolutePath());
						line = reader.readLine();
						reader.close();
					} catch (Exception e) {
						Logger.error(String.format("Cannot read command in %s!", f.getAbsolutePath()), e);
						continue;
					}
					if (line != null) {
						try {
							perform(line);
							f.delete();
						} catch (Exception e) {
							f.renameTo(new File(f.getAbsoluteFile() + ".err"));
							Logger.error(String.format("Cannot execute command in %s!", f.getAbsolutePath()), e);
						}
					}
				}
				Object o = new Object();
				synchronized (o) {
					o.wait(2000);
				}
			}
		} catch (Exception e) {
			Logger.error(BioinfException.convert(e));
		}
	}

	/**
	 * Perform the action on the comamnd on the read line.
	 */
	public abstract void perform(String line) throws BioinfException;
}
