package de.bioinf.utils;

/**
 * This class provides some nice mechanism to abstract from loading/writing parameters.
 * 
 * @author ng
 */
public abstract class PersistentParams<K, V> extends Params<K, V> {
	private String filename = null;

	public PersistentParams() {
		super();
	}

	public PersistentParams(String filename) throws BioinfException {
		super();
		load(filename);
	}

	public String getFilename() {
		return filename;
	}

	/**
	 * Creates a key type for the keyed string value (read from the file)
	 */
	public abstract K getKeyObject(String key);

	/**
	 * Creates a value type for the String s (read from the file).
	 * 
	 * @return The value object or -if not possible- null.
	 */
	public abstract V getValueObject(String s) throws BioinfException;

	/**
	 * Loads the data from a file. The actual data is overwritten.
	 */
	public void load(String filename) throws BioinfException {
		LineReader reader = new LineReader(this.filename = filename);
		load(reader);
		reader.close();
	}

	/**
	 * Loads the data using a LineReader.
	 */
	public void load(LineReader reader) throws BioinfException {
		clear();
		try {
			dirty = true;
			String line = null;
			while ((line = reader.readLine()) != null) {
				int index = line.indexOf('=');
				if (index != -1) {
					String skey = line.substring(0, index).trim();
					String svalue = line.substring(index + 1).trim();
					put(getKeyObject(skey), getValueObject(svalue));
				}
			}
			dirty = false;
		} catch (Exception e) {
			throw BioinfException.convert(e);
		}
	}

	/**
	 * Save the actual data back (only if it is changed).
	 */
	public final void save() throws BioinfException {
		if (dirty) {
			LineWriter out = new LineWriter(filename);
			K[] keys = getSortedKeys();
			for (int i = 0; i < keys.length; i++) {
				out.writeLine(keys[i].toString() + "=" + toString(getValue(keys[i])));
			}
			out.close();
			dirty = false;
		}
	}

	/**
	 * Save on file - in every case, if the parameters are changed or not. The filename is taken as the actual filename,
	 * so future savings are done into this file!
	 */
	public final void save(String filename) throws BioinfException {
		this.filename = filename;
		dirty = true;
		save();
	}

	/**
	 * Used e.g. in save(). It is the opposite of getValueObject()
	 */
	public abstract String toString(V value);
}
