package de.bioinf.utils;

import java.io.File;
import java.net.URL;
import java.util.Date;

/**
 * A class specializing the location of data, which can be either a name of a file, a directory or an URL.
 * 
 * @author ng
 */
public class Source {
	private String location = null;
	private long lastModified = 0;
	private URL url = null;
	private File file = null;

	/**
	 * Creates a source object and sets the last modification date.
	 */
	public Source(String location) {
		this(location, 0);
	}

	/**
	 * Create a Source object, but sets the last modified date to the specified value.
	 * 
	 * @param lastModified - the modification date. If greater 0 this is taken otherwise it s read from the specified
	 *        location.
	 */
	public Source(String location, long lastModified) {
		super();
		this.location = location;
		try {
			url = new URL(this.location);
		} catch (Exception e) {
			file = new File(this.location);
		}
		try {
			if (lastModified > 0) {
				this.lastModified = lastModified;
			} else {
				this.lastModified = url != null ? url.openConnection().getLastModified() : file.lastModified();
			}
		} catch (Exception e) {
		}
	}

	/**
	 * Returns the location as a string.
	 */
	public String getLocation() {
		return location;
	}

	/**
	 * Returns the modification time of the file
	 */
	public long getLastModified() {
		return lastModified;
	}

	/**
	 * Returns a line reader with optional progress "display" - that means set this to true call Logger.info() every
	 * second.
	 */
	public LineReader getReader(boolean progress) throws BioinfException {
		try {
			return isUrl() ? new LineReader(url, progress) : new LineReader(file, progress);
		} catch (Exception e) {
			throw BioinfException.convert(e);
		}
	}

	/**
	 * Is this location an url?
	 */
	public boolean isUrl() {
		return url != null;
	}

	/**
	 * Return true if the location and modification time are the same.
	 */
	public boolean equals(Object o) {
		if (o == null || !(o instanceof Source)) {
			return false;
		}
		Source source = (Source) o;
		return location.equals(source.location) && getLastModified() == source.getLastModified();
	}

	public String toString() {
		return location + "[" + Utils.DATE_FORMAT.format(new Date(lastModified)) + "]";
	}
}
