package de.bioinf.utils;

import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.io.InputStream;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.StringTokenizer;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

/**
 * Class for systemwide utilities.
 * 
 * @author ng
 */
public class Utils {
	public static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy.MM.dd HH:mm");

	/**
	 * A file filter that filters using a matcher and before matching is done the prefix is removed from the filename.
	 */
	static class CpFileFilter implements FileFilter {
		Matcher matcher = null;
		String prefix = null;

		public CpFileFilter(Matcher m, String prefix) {
			this.matcher = m;
			this.prefix = prefix;
		}

		public boolean accept(File f) {
			return f.isDirectory() || matcher.matches(removePrefix(f.getAbsolutePath(), prefix));
		}
	}

	static class FullFileFilter implements FilenameFilter {
		String pattern = null;

		public FullFileFilter(String pattern) {
			this.pattern = pattern;
		}

		public boolean accept(File f, String name) {
			return name.matches(pattern);
		}
	}

	// Cannot instantiate this class.
	private Utils() {
	}

	/**
	 * Returns the user specific configuration directory. This directory is called ".bioinf" and placed under the user's
	 * home directory.
	 */
	public static String getConfDirectory() {
		return getHomeDirectory() + "/.bioinf";
	}

	/**
	 * Returns the classname of the object without the package specification.
	 */
	public static String getName(Object o) {
		String name = o.getClass().getName();
		int index = name.lastIndexOf('.');
		return index != -1 ? name.substring(index + 1) : name;
	}

	/**
	 * Splits the full filename qualification (this may even be an URL!) into a prefix (e.g. a directory) and the
	 * filename itself, which in fact is an array with 2 elements. If there is no directory/prefix, "/" will be
	 * returned.
	 */
	public static String[] splitFilename(String fullname) {
		String[] dirfilename = new String[] { ".", fullname };
		int index = fullname.lastIndexOf('/');
		if (index < 0) {
			index = fullname.lastIndexOf('\\');
		}
		if (index >= 0) {
			dirfilename[0] = fullname.substring(0, index);
			dirfilename[1] = fullname.substring(index + 1);
		}
		return dirfilename;
	}

	/**
	 * Returns the filename without the path
	 */
	public static String getFilename(Source source) {
		return getFilename(source.getLocation());
	}

	/**
	 * Returns the filename without the path
	 */
	public static String getFilename(String fullname) {
		return splitFilename(fullname)[1];
	}

	/**
	 * Returns the suffix, or null if no suffix present. A suffix is separated from the rest by a ".".
	 */
	public static String getSuffix(String filename) {
		int index = filename.lastIndexOf(".");
		if (index >= 0) {
			return filename.substring(index + 1);
		}
		return null;
	}

	/**
	 * Checks, if the source end with one of the specified suffixes and returns true or false.
	 */
	public static boolean checkSuffixes(String filename, String[] suffixes) {
		String suffix = getSuffix(filename);
		if (suffix != null) {
			for (String s : suffixes) {
				if (s.equals(suffix)) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Returns an array of filenames matches the fullname specification, which may contain the wildcards "." and "*".
	 */
	public static String[] getFullFilenames(String fullname) {
		String[] parts = splitFilename(fullname);
		parts[1] = parts[1].replace(".", "\\.").replace("*", ".*");
		String[] filenames = new File(parts[0]).list(new FullFileFilter(parts[1]));
		for (int i = 0; i < filenames.length; i++) {
			filenames[i] = parts[0] + "/" + filenames[i];
		}
		return filenames;
	}

	/**
	 * Returns the content of a file that is accessible through the classpath. It does NOT get contents somewhere in the
	 * filesystem!
	 */
	public static byte[] getClassPathContent(String filename) throws BioinfException {
		try {
			List<String> cpelems = getClassPathElements();
			for (int i = 0; i < cpelems.size(); i++) {
				String cpelem = cpelems.get(i);
				File file = new File(cpelem);
				if (file.exists()) {
					if (file.isFile()) {
						if (file.getName().endsWith(".jar")) {
							JarFile jfile = new JarFile(file);
							Enumeration<JarEntry> entries = jfile.entries();
							while (entries.hasMoreElements()) {
								JarEntry entry = entries.nextElement();
								if (filename.equals(entry.getName())) {
									InputStream in = jfile.getInputStream(entry);
									return read(in, (int) entry.getSize());
								}
							}
						}
					} else if (file.isDirectory()) {
						file = new File(cpelem + File.separatorChar + filename);
						if (file.isFile()) {
							FileInputStream in = new FileInputStream(file);
							return read(in, (int) in.getChannel().size());
						}
					}
				}
			}
			return null;
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage("Cannot read file '" + filename + "' in classpath!");
		}
	}

	/**
	 * Returns the list of classpath elements (separated by : or ;)
	 */
	public static ArrayList<String> getClassPathElements() {
		ArrayList<String> elems = new ArrayList<String>();
		StringTokenizer tok = new StringTokenizer(System.getProperty("java.class.path"), ":;");
		while (tok.hasMoreTokens()) {
			elems.add(tok.nextToken());
		}
		return elems;
	}

	/**
	 * Returns the filenames that match a pattern and can be found somewhere in the classpath. The returned filenames
	 * are "relative", that means they are of kind: "conf/xyz", but without the classpath-prefix!
	 */
	public static String[] getClassPathFiles(Matcher m) throws BioinfException {
		try {
			HashSet<String> filenames = new HashSet<String>();
			List<String> cpelems = getClassPathElements();
			for (int i = 0; i < cpelems.size(); i++) {
				String cpelem = cpelems.get(i);
				File file = new File(cpelem);
				if (file.exists()) {
					if (file.isFile()) {
						if (file.getName().endsWith(".jar")) {
							JarFile jfile = new JarFile(file);
							Enumeration<JarEntry> entries = jfile.entries();
							while (entries.hasMoreElements()) {
								JarEntry entry = entries.nextElement();
								if (m.matches(entry.getName())) {
									filenames.add(entry.getName());
								}
							}
						}
					} else if (file.isDirectory()) {
						ArrayList<File> files = new ArrayList<File>();
						fillFiles(file, m, cpelem, files);
						String prefix = cpelem + "/";
						for (File f : files) {
							filenames.add(removePrefix(f.getAbsolutePath(), prefix));
						}
					}
				}
			}
			String[] ss = filenames.toArray(new String[filenames.size()]);
			Arrays.sort(ss);
			return ss;
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage("Cannot find files for pattern '" + m + "' in classpath!");
		}
	}

	/**
	 * Returns the home directory.
	 */
	public static String getHomeDirectory() {
		return System.getProperty("user.home");
	}

	/**
	 * Returns the current time as a string (in the form 2007-01-04 14:50)
	 */
	public static String getCurrentTime() {
		return DATE_FORMAT.format(new Date(System.currentTimeMillis()));
	}

	/**
	 * Is the string an Integer/int value?
	 */
	public static boolean isInteger(String s) {
		try {
			Integer.parseInt(s);
			return true;
		} catch (Exception e) {
			return false;
		}
	}

	public static String[] toArray(String s, String delims, boolean withDelims) {
		ArrayList<String> a = new ArrayList<String>();
		int pos = 0;
		while (pos < s.length()) {
			// Ignore the delimiters
			while (pos < s.length() && delims.indexOf(s.charAt(pos)) >= 0) {
				if (withDelims) {
					a.add(s.substring(pos, pos + 1));
				}
				pos++;
			}
			// Look for the end of the part and add it.
			int start = pos;
			while (pos < s.length() && delims.indexOf(s.charAt(pos)) < 0) {
				pos++;
			}
			a.add(s.substring(start, pos));
		}
		return a.toArray(new String[a.size()]);
	}

	/**
	 * Cuts a string into an array of substrings with the given delimiters. The delimiters are not returned!
	 */
	public static String[] toArray(String s, String delims) {
		return toArray(s, delims, false);
	}

	/**
	 * Cuts a string into an array of substrings with the delimiters " \t\n".
	 */
	public static String[] toArray(String s) {
		return toArray(s, " \t\n");
	}

	/**
	 * Get the files of an URL (the URL should be written as
	 * "http://www.compgen.uni-muenster.de/tools/tint/data/.ADIR?pattern=*.rm", the extension .ADIR is neccessary!
	 */
	public static String[] getFiles(URL url) {
		try {
			LineReader reader = new LineReader(url);
			String line = null;
			ArrayList<String> lines = new ArrayList<String>();
			while ((line = reader.readLine()) != null) {
				lines.add(line);
			}
			return (String[]) lines.toArray(new String[lines.size()]);
		} catch (Exception e) {
			Logger.error(String.format("Cannot read %s!", url.toString()));
			return new String[0];
		}
	}

	/**
	 * A simple "abbreviaton" for sort() and return call. Makes code smaller
	 */
	public static Object[] sort(Object[] os) {
		Arrays.sort(os);
		return os;
	}

	// **
	// Private
	// **

	private static void fillFiles(File file, Matcher m, String cpelem, ArrayList<File> files) {
		if (file.isFile()) {
			// A file can be direct added - the initial call is made with a
			// directory and calls afterwards are made through a FileFilter.
			files.add(file);
		} else if (file.isDirectory()) {
			File[] fs = file.listFiles(new CpFileFilter(m, cpelem));
			for (int i = 0; i < fs.length; i++) {
				fillFiles(fs[i], m, cpelem, files);
			}
		}
	}

	private static String removePrefix(String s, String prefix) {
		return s.startsWith(prefix) ? s.substring(prefix.length()) : s;
	}

	/**
	 * Reads count bytes from a stream. On error it returns null.
	 */
	private static byte[] read(InputStream in, int count) {
		try {
			byte[] bytes = new byte[count];
			in.read(bytes);
			in.close();
			return bytes;
		} catch (Exception e) {
			return null;
		}
	}
}
